from __future__ import annotations

import typing
from typing import Any, Final

import zigpy.types as t


class CurrentPowerMode(t.enum4):
    RxOnSyncedWithNodeDesc = 0b0000
    RxOnPeriodically = 0b0001
    RxOnWhenStimulated = 0b0010


class PowerSources(t.bitmap4):
    MainsPower = 0b0001
    RechargeableBattery = 0b0010
    DisposableBattery = 0b0100
    Reserved = 0b1000


class PowerSourceLevel(t.enum4):
    Critical = 0b0000
    Percent33 = 0b0100
    Percent66 = 0b1000
    Percent100 = 0b1100


class PowerDescriptor(t.Struct):
    current_power_mode: CurrentPowerMode
    available_power_sources: PowerSources
    current_power_source: PowerSources
    current_power_source_level: PowerSourceLevel

    # Kept for backwards compatibility
    CurrentPowerMode: Final = CurrentPowerMode
    PowerSources: Final = PowerSources
    PowerSourceLevel: Final = PowerSourceLevel


class SimpleDescriptor(t.Struct):
    endpoint: t.uint8_t
    profile: t.uint16_t
    device_type: t.uint16_t
    device_version: t.uint8_t
    input_clusters: t.LVList[t.uint16_t, t.uint8_t]
    output_clusters: t.LVList[t.uint16_t, t.uint8_t]


class SizePrefixedSimpleDescriptor(SimpleDescriptor):
    def serialize(self):
        data = super().serialize()
        return len(data).to_bytes(1, "little") + data

    @classmethod
    def deserialize(cls, data):
        if not data or data[0] == 0:
            return None, data[1:]
        return super().deserialize(data[1:])


class LogicalType(t.enum3):
    Coordinator = 0b000
    Router = 0b001
    EndDevice = 0b010


class MACCapabilityFlags(t.bitmap8):
    NONE = 0

    AlternatePanCoordinator = 0b00000001
    FullFunctionDevice = 0b00000010
    MainsPowered = 0b00000100
    RxOnWhenIdle = 0b00001000
    SecurityCapable = 0b01000000
    AllocateAddress = 0b10000000


class FrequencyBand(t.bitmap5):
    Freq868MHz = 0b00001
    Freq902MHz = 0b00100
    Freq2400MHz = 0b01000


class DescriptorCapability(t.bitmap8):
    NONE = 0

    ExtendedActiveEndpointListAvailable = 0b00000001
    ExtendedSimpleDescriptorListAvailable = 0b00000010


class NodeDescriptor(t.Struct):
    logical_type: LogicalType
    complex_descriptor_available: t.uint1_t
    user_descriptor_available: t.uint1_t
    reserved: t.uint3_t

    aps_flags: t.uint3_t
    frequency_band: FrequencyBand

    mac_capability_flags: MACCapabilityFlags
    manufacturer_code: t.uint16_t
    maximum_buffer_size: t.uint8_t
    maximum_incoming_transfer_size: t.uint16_t
    server_mask: t.uint16_t
    maximum_outgoing_transfer_size: t.uint16_t
    descriptor_capability_field: DescriptorCapability

    def __new__(cls, *args, **kwargs):
        # Old style constructor
        if len(args) == 9 or "byte1" in kwargs or "byte2" in kwargs:
            return cls._old_constructor(*args, **kwargs)

        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def _old_constructor(
        cls: NodeDescriptor,
        byte1: t.uint8_t = None,
        byte2: t.uint8_t = None,
        mac_capability_flags: MACCapabilityFlags = None,
        manufacturer_code: t.uint16_t = None,
        maximum_buffer_size: t.uint8_t = None,
        maximum_incoming_transfer_size: t.uint16_t = None,
        server_mask: t.uint16_t = None,
        maximum_outgoing_transfer_size: t.uint16_t = None,
        descriptor_capability_field: t.uint8_t = None,
    ) -> NodeDescriptor:
        logical_type = None
        complex_descriptor_available = None
        user_descriptor_available = None
        reserved = None

        if byte1 is not None:
            bits, _ = t.Bits.deserialize(bytes([byte1]))
            logical_type, bits = LogicalType.from_bits(bits)
            complex_descriptor_available, bits = t.uint1_t.from_bits(bits)
            user_descriptor_available, bits = t.uint1_t.from_bits(bits)
            reserved, bits = t.uint3_t.from_bits(bits)

            assert not bits

        aps_flags = None
        frequency_band = None

        if byte2 is not None:
            bits, _ = t.Bits.deserialize(bytes([byte2]))
            aps_flags, bits = t.uint3_t.from_bits(bits)
            frequency_band, bits = cls.FrequencyBand.from_bits(bits)

            assert not bits

        return cls(  # type:ignore[operator]
            logical_type=logical_type,
            complex_descriptor_available=complex_descriptor_available,
            user_descriptor_available=user_descriptor_available,
            reserved=reserved,
            aps_flags=aps_flags,
            frequency_band=frequency_band,
            mac_capability_flags=mac_capability_flags,
            manufacturer_code=manufacturer_code,
            maximum_buffer_size=maximum_buffer_size,
            maximum_incoming_transfer_size=maximum_incoming_transfer_size,
            server_mask=server_mask,
            maximum_outgoing_transfer_size=maximum_outgoing_transfer_size,
            descriptor_capability_field=descriptor_capability_field,
        )

    @property
    def is_end_device(self) -> bool | None:
        if self.logical_type is None:
            return None

        return self.logical_type == LogicalType.EndDevice

    @property
    def is_router(self) -> bool | None:
        if self.logical_type is None:
            return None

        return self.logical_type == LogicalType.Router

    @property
    def is_coordinator(self) -> bool | None:
        if self.logical_type is None:
            return None

        return self.logical_type == LogicalType.Coordinator

    @property
    def is_alternate_pan_coordinator(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(
            self.mac_capability_flags & self.MACCapabilityFlags.AlternatePanCoordinator
        )

    @property
    def is_full_function_device(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(
            self.mac_capability_flags & self.MACCapabilityFlags.FullFunctionDevice
        )

    @property
    def is_mains_powered(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(self.mac_capability_flags & self.MACCapabilityFlags.MainsPowered)

    @property
    def is_receiver_on_when_idle(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(self.mac_capability_flags & self.MACCapabilityFlags.RxOnWhenIdle)

    @property
    def is_security_capable(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(self.mac_capability_flags & self.MACCapabilityFlags.SecurityCapable)

    @property
    def allocate_address(self) -> bool | None:
        if self.mac_capability_flags is None:
            return None

        return bool(self.mac_capability_flags & self.MACCapabilityFlags.AllocateAddress)

    # Kept for backwards compatibility
    FrequencyBand: Final = FrequencyBand
    MACCapabilityFlags: Final = MACCapabilityFlags
    DescriptorCapability: Final = DescriptorCapability


class MultiAddress(t.Struct):
    """Used for binds, represents an IEEE+endpoint or NWK address"""

    addrmode: t.uint8_t
    nwk: t.uint16_t = t.StructField(requires=lambda s: s.addrmode == 0x01)
    ieee: t.EUI64 = t.StructField(requires=lambda s: s.addrmode == 0x03)
    endpoint: t.uint8_t = t.StructField(requires=lambda s: s.addrmode == 0x03)

    @classmethod
    def deserialize(cls, data):
        r, data = super().deserialize(data)

        if r.addrmode not in (0x01, 0x03):
            raise ValueError("Invalid MultiAddress - unknown address mode")

        return r, data

    def serialize(self):
        if self.addrmode not in (0x01, 0x03):
            raise ValueError("Invalid MultiAddress - unknown address mode")

        return super().serialize()


class DeviceType(t.enum2):
    Coordinator = 0x0
    Router = 0x1
    EndDevice = 0x2
    Unknown = 0x3


class RxOnWhenIdle(t.enum2):
    Off = 0x0
    On = 0x1
    Unknown = 0x2


class Relationship(t.enum3):
    Parent = 0x0
    Child = 0x1
    Sibling = 0x2
    NoneOfTheAbove = 0x3
    PreviousChild = 0x4


class PermitJoins(t.enum2):
    NotAccepting = 0x0
    Accepting = 0x1
    Unknown = 0x2


class Neighbor(t.Struct):
    """Neighbor Descriptor"""

    # Backwards-compatible alternate spelling
    RelationShip = Relationship

    extended_pan_id: t.ExtendedPanId
    ieee: t.EUI64
    nwk: t.NWK

    device_type: DeviceType
    rx_on_when_idle: RxOnWhenIdle
    relationship: Relationship
    reserved1: t.uint1_t

    permit_joining: PermitJoins
    reserved2: t.uint6_t

    depth: t.uint8_t
    lqi: t.uint8_t

    @classmethod
    def _parse_packed(cls, packed: t.uint8_t) -> dict[str, typing.Any]:
        data = 18 * b"\x00" + t.uint16_t(packed).serialize() + 3 * b"\x00"
        tmp_neighbor, _ = cls.deserialize(data)

        return {
            "device_type": tmp_neighbor.device_type,
            "rx_on_when_idle": tmp_neighbor.rx_on_when_idle,
            "relationship": tmp_neighbor.relationship,
            "reserved1": tmp_neighbor.reserved1,
        }

    # Kept for backwards compatibility
    PermitJoins: Final = PermitJoins
    DeviceType: Final = DeviceType
    RxOnWhenIdle: Final = RxOnWhenIdle
    Relationship: Final = Relationship


class Neighbors(t.Struct):
    """Mgmt_Lqi_rsp"""

    Entries: t.uint8_t
    StartIndex: t.uint8_t
    NeighborTableList: t.LVList[Neighbor, t.uint8_t]


class RouteStatus(t.enum3):
    """Route descriptor route status."""

    Active = 0x00
    Discovery_Underway = 0x01
    Discovery_Failed = 0x02
    Inactive = 0x03
    Validation_Underway = 0x04
    Reserved_5 = 0x05
    Reserved_6 = 0x06
    Reserved_7 = 0x07


class Route(t.Struct):
    """Route Descriptor"""

    DstNWK: t.NWK
    RouteStatus: RouteStatus

    # Whether the device is a memory constrained concentrator.
    MemoryConstrained: t.uint1_t

    # The destination is a concentrator that issued a many-to-one request.
    ManyToOne: t.uint1_t

    # A route record command frame should be sent to the destination prior to the next
    # data packet.
    RouteRecordRequired: t.uint1_t
    Reserved: t.uint2_t

    NextHop: t.NWK


class Routes(t.Struct):
    Entries: t.uint8_t
    StartIndex: t.uint8_t
    RoutingTableList: t.LVList[Route, t.uint8_t]


CHANNEL_CHANGE_REQ = 0xFE
CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ = 0xFF


class NwkUpdate(t.Struct):
    CHANNEL_CHANGE_REQ = CHANNEL_CHANGE_REQ
    CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ = CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ

    ScanChannels: t.Channels
    ScanDuration: t.uint8_t
    ScanCount: t.uint8_t = t.StructField(requires=lambda s: s.ScanDuration <= 0x05)
    nwkUpdateId: t.uint8_t = t.StructField(  # noqa: N815
        requires=lambda s: s.ScanDuration
        in (CHANNEL_CHANGE_REQ, CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ)
    )
    nwkManagerAddr: t.NWK = t.StructField(  # noqa: N815
        requires=lambda s: s.ScanDuration == CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ
    )


class Binding(t.Struct):
    SrcAddress: t.EUI64
    SrcEndpoint: t.uint8_t
    ClusterId: t.uint16_t
    DstAddress: MultiAddress


class AddrRequestType(t.enum8):
    Single = 0x00
    Extended = 0x01


class Status(t.enum8):
    # The requested operation or transmission was completed successfully.
    SUCCESS = 0x00
    # The supplied request type was invalid.
    INV_REQUESTTYPE = 0x80
    # The requested device did not exist on a device following a child
    # descriptor request to a parent.
    DEVICE_NOT_FOUND = 0x81
    # The supplied endpoint was equal to = 0x00 or between 0xf1 and 0xff.
    INVALID_EP = 0x82
    # The requested endpoint is not described by a simple descriptor.
    NOT_ACTIVE = 0x83
    # The requested optional feature is not supported on the target device.
    NOT_SUPPORTED = 0x84
    # A timeout has occurred with the requested operation.
    TIMEOUT = 0x85
    # The end device bind request was unsuccessful due to a failure to match
    # any suitable clusters.
    NO_MATCH = 0x86
    # The unbind request was unsuccessful due to the coordinator or source
    # device not having an entry in its binding table to unbind.
    NO_ENTRY = 0x88
    # A child descriptor was not available following a discovery request to a
    # parent.
    NO_DESCRIPTOR = 0x89
    # The device does not have storage space to support the requested
    # operation.
    INSUFFICIENT_SPACE = 0x8A
    # The device is not in the proper state to support the requested operation.
    NOT_PERMITTED = 0x8B
    # The device does not have table space to support the operation.
    TABLE_FULL = 0x8C
    # The permissions configuration table on the target indicates that the
    # request is not authorized from this device.
    NOT_AUTHORIZED = 0x8D

    @classmethod
    def _missing_(cls, value):
        chained = t.APSStatus(value)
        status = cls._member_type_.__new__(cls, chained.value)
        status._name_ = chained.name
        status._value_ = value
        return status


NWK = ("NWKAddr", t.NWK)
NWKI = ("NWKAddrOfInterest", t.NWK)
IEEE = ("IEEEAddr", t.EUI64)
STATUS = ("Status", Status)


class ZDOCmd(t.enum16):
    # Device and Service Discovery Server Requests
    NWK_addr_req = 0x0000
    IEEE_addr_req = 0x0001
    Node_Desc_req = 0x0002
    Power_Desc_req = 0x0003
    Simple_Desc_req = 0x0004
    Active_EP_req = 0x0005
    Match_Desc_req = 0x0006
    Complex_Desc_req = 0x0010
    User_Desc_req = 0x0011
    Discovery_Cache_req = 0x0012
    Device_annce = 0x0013
    User_Desc_set = 0x0014
    System_Server_Discovery_req = 0x0015
    Discovery_store_req = 0x0016
    Node_Desc_store_req = 0x0017
    Active_EP_store_req = 0x0019
    Simple_Desc_store_req = 0x001A
    Remove_node_cache_req = 0x001B
    Find_node_cache_req = 0x001C
    Extended_Simple_Desc_req = 0x001D
    Extended_Active_EP_req = 0x001E
    Parent_annce = 0x001F
    #  Bind Management Server Services Responses
    End_Device_Bind_req = 0x0020
    Bind_req = 0x0021
    Unbind_req = 0x0022
    # Network Management Server Services Requests
    # ... TODO optional stuff ...
    Mgmt_Lqi_req = 0x0031
    Mgmt_Rtg_req = 0x0032
    Mgmt_Bind_req = 0x0033
    # ... TODO optional stuff ...
    Mgmt_Leave_req = 0x0034
    Mgmt_Permit_Joining_req = 0x0036
    Mgmt_NWK_Update_req = 0x0038
    # ... TODO optional stuff ...

    # Responses
    # Device and Service Discovery Server Responses
    NWK_addr_rsp = 0x8000
    IEEE_addr_rsp = 0x8001
    Node_Desc_rsp = 0x8002
    Power_Desc_rsp = 0x8003
    Simple_Desc_rsp = 0x8004
    Active_EP_rsp = 0x8005
    Match_Desc_rsp = 0x8006
    Complex_Desc_rsp = 0x8010
    User_Desc_rsp = 0x8011
    Discovery_Cache_rsp = 0x8012
    User_Desc_conf = 0x8014
    System_Server_Discovery_rsp = 0x8015
    Discovery_Store_rsp = 0x8016
    Node_Desc_store_rsp = 0x8017
    Power_Desc_store_rsp = 0x8018
    Active_EP_store_rsp = 0x8019
    Simple_Desc_store_rsp = 0x801A
    Remove_node_cache_rsp = 0x801B
    Find_node_cache_rsp = 0x801C
    Extended_Simple_Desc_rsp = 0x801D
    Extended_Active_EP_rsp = 0x801E
    Parent_annce_rsp = 0x801F
    #  Bind Management Server Services Responses
    End_Device_Bind_rsp = 0x8020
    Bind_rsp = 0x8021
    Unbind_rsp = 0x8022
    # ... TODO optional stuff ...
    # Network Management Server Services Responses
    Mgmt_Lqi_rsp = 0x8031
    Mgmt_Rtg_rsp = 0x8032
    Mgmt_Bind_rsp = 0x8033
    # ... TODO optional stuff ...
    Mgmt_Leave_rsp = 0x8034
    Mgmt_Permit_Joining_rsp = 0x8036
    # ... TODO optional stuff ...
    Mgmt_NWK_Update_rsp = 0x8038


CLUSTERS: dict[ZDOCmd, tuple[tuple[str, Any], ...]] = {
    # Device and Service Discovery Server Requests
    ZDOCmd.NWK_addr_req: (
        IEEE,
        ("RequestType", AddrRequestType),
        ("StartIndex", t.uint8_t),
    ),
    ZDOCmd.IEEE_addr_req: (
        NWKI,
        ("RequestType", AddrRequestType),
        ("StartIndex", t.uint8_t),
    ),
    ZDOCmd.Node_Desc_req: (NWKI,),
    ZDOCmd.Power_Desc_req: (NWKI,),
    ZDOCmd.Simple_Desc_req: (NWKI, ("EndPoint", t.uint8_t)),
    ZDOCmd.Active_EP_req: (NWKI,),
    ZDOCmd.Match_Desc_req: (
        NWKI,
        ("ProfileID", t.uint16_t),
        ("InClusterList", t.LVList[t.uint16_t, t.uint8_t]),
        ("OutClusterList", t.LVList[t.uint16_t, t.uint8_t]),
    ),
    # ZDO.Complex_Desc_req: (NWKI, ),
    ZDOCmd.User_Desc_req: (NWKI,),
    ZDOCmd.Discovery_Cache_req: (NWK, IEEE),
    ZDOCmd.Device_annce: (NWK, IEEE, ("Capability", t.uint8_t)),
    ZDOCmd.User_Desc_set: (
        NWKI,
        ("UserDescriptor", t.CharacterString),
    ),  # Really a string
    ZDOCmd.System_Server_Discovery_req: (("ServerMask", t.uint16_t),),
    ZDOCmd.Discovery_store_req: (
        NWK,
        IEEE,
        ("NodeDescSize", t.uint8_t),
        ("PowerDescSize", t.uint8_t),
        ("ActiveEPSize", t.uint8_t),
        ("SimpleDescSizeList", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    ZDOCmd.Node_Desc_store_req: (NWK, IEEE, ("NodeDescriptor", NodeDescriptor)),
    ZDOCmd.Active_EP_store_req: (
        NWK,
        IEEE,
        ("ActiveEPList", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    ZDOCmd.Simple_Desc_store_req: (
        NWK,
        IEEE,
        ("SimpleDescriptor", SizePrefixedSimpleDescriptor),
    ),
    ZDOCmd.Remove_node_cache_req: (NWK, IEEE),
    ZDOCmd.Find_node_cache_req: (NWK, IEEE),
    ZDOCmd.Extended_Simple_Desc_req: (
        NWKI,
        ("EndPoint", t.uint8_t),
        ("StartIndex", t.uint8_t),
    ),
    ZDOCmd.Extended_Active_EP_req: (NWKI, ("StartIndex", t.uint8_t)),
    ZDOCmd.Parent_annce: (("Children", t.LVList[t.EUI64, t.uint8_t]),),
    #  Bind Management Server Services Responses
    ZDOCmd.End_Device_Bind_req: (
        ("BindingTarget", t.uint16_t),
        ("SrcAddress", t.EUI64),
        ("SrcEndpoint", t.uint8_t),
        ("ProfileID", t.uint8_t),
        ("InClusterList", t.LVList[t.uint8_t, t.uint8_t]),
        ("OutClusterList", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    ZDOCmd.Bind_req: (
        ("SrcAddress", t.EUI64),
        ("SrcEndpoint", t.uint8_t),
        ("ClusterID", t.uint16_t),
        ("DstAddress", MultiAddress),
    ),
    ZDOCmd.Unbind_req: (
        ("SrcAddress", t.EUI64),
        ("SrcEndpoint", t.uint8_t),
        ("ClusterID", t.uint16_t),
        ("DstAddress", MultiAddress),
    ),
    # Network Management Server Services Requests
    # ... TODO optional stuff ...
    ZDOCmd.Mgmt_Lqi_req: (("StartIndex", t.uint8_t),),
    ZDOCmd.Mgmt_Rtg_req: (("StartIndex", t.uint8_t),),
    ZDOCmd.Mgmt_Bind_req: (("StartIndex", t.uint8_t),),
    # ... TODO optional stuff ...
    ZDOCmd.Mgmt_Leave_req: (("DeviceAddress", t.EUI64), ("Options", t.bitmap8)),
    ZDOCmd.Mgmt_Permit_Joining_req: (
        ("PermitDuration", t.uint8_t),
        ("TC_Significant", t.Bool),
    ),
    ZDOCmd.Mgmt_NWK_Update_req: (("NwkUpdate", NwkUpdate),),
    # ... TODO optional stuff ...
    # Responses
    # Device and Service Discovery Server Responses
    ZDOCmd.NWK_addr_rsp: (
        STATUS,
        IEEE,
        NWK,
        ("NumAssocDev", t.Optional(t.uint8_t)),
        ("StartIndex", t.Optional(t.uint8_t)),
        ("NWKAddressAssocDevList", t.Optional(t.List[t.NWK])),
    ),
    ZDOCmd.IEEE_addr_rsp: (
        STATUS,
        IEEE,
        NWK,
        ("NumAssocDev", t.Optional(t.uint8_t)),
        ("StartIndex", t.Optional(t.uint8_t)),
        ("NWKAddrAssocDevList", t.Optional(t.List[t.NWK])),
    ),
    ZDOCmd.Node_Desc_rsp: (
        STATUS,
        NWKI,
        ("NodeDescriptor", t.Optional(NodeDescriptor)),
    ),
    ZDOCmd.Power_Desc_rsp: (
        STATUS,
        NWKI,
        ("PowerDescriptor", t.Optional(PowerDescriptor)),
    ),
    ZDOCmd.Simple_Desc_rsp: (
        STATUS,
        NWKI,
        ("SimpleDescriptor", t.Optional(SizePrefixedSimpleDescriptor)),
    ),
    ZDOCmd.Active_EP_rsp: (
        STATUS,
        NWKI,
        ("ActiveEPList", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    ZDOCmd.Match_Desc_rsp: (
        STATUS,
        NWKI,
        ("MatchList", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    # ZDO.Complex_Desc_rsp: (
    #     STATUS,
    #     NWKI,
    #     ('Length', t.uint8_t),
    #     ('ComplexDescriptor', t.Optional(ComplexDescriptor)),
    # ),
    ZDOCmd.User_Desc_rsp: (
        STATUS,
        NWKI,
        ("UserDescriptor", t.CharacterString),
    ),
    ZDOCmd.Discovery_Cache_rsp: (STATUS,),
    ZDOCmd.User_Desc_conf: (STATUS, NWKI),
    ZDOCmd.System_Server_Discovery_rsp: (STATUS, ("ServerMask", t.uint16_t)),
    ZDOCmd.Discovery_Store_rsp: (STATUS,),
    ZDOCmd.Node_Desc_store_rsp: (STATUS,),
    ZDOCmd.Power_Desc_store_rsp: (STATUS, IEEE, ("PowerDescriptor", PowerDescriptor)),
    ZDOCmd.Active_EP_store_rsp: (STATUS,),
    ZDOCmd.Simple_Desc_store_rsp: (STATUS,),
    ZDOCmd.Remove_node_cache_rsp: (STATUS,),
    ZDOCmd.Find_node_cache_rsp: (("CacheNWKAddr", t.EUI64), NWK, IEEE),
    ZDOCmd.Extended_Simple_Desc_rsp: (
        STATUS,
        NWK,
        ("Endpoint", t.uint8_t),
        ("AppInputClusterCount", t.uint8_t),
        ("AppOutputClusterCount", t.uint8_t),
        ("StartIndex", t.uint8_t),
        ("AppClusterList", t.Optional(t.List[t.uint16_t])),
    ),
    ZDOCmd.Extended_Active_EP_rsp: (
        STATUS,
        NWKI,
        ("ActiveEPCount", t.uint8_t),
        ("StartIndex", t.uint8_t),
        ("ActiveEPList", t.List[t.uint8_t]),
    ),
    ZDOCmd.Parent_annce_rsp: (STATUS, ("Children", t.LVList[t.EUI64, t.uint8_t])),
    #  Bind Management Server Services Responses
    ZDOCmd.End_Device_Bind_rsp: (STATUS,),
    ZDOCmd.Bind_rsp: (STATUS,),
    ZDOCmd.Unbind_rsp: (STATUS,),
    # ... TODO optional stuff ...
    # Network Management Server Services Responses
    ZDOCmd.Mgmt_Lqi_rsp: (STATUS, ("Neighbors", t.Optional(Neighbors))),
    ZDOCmd.Mgmt_Rtg_rsp: (STATUS, ("Routes", t.Optional(Routes))),
    ZDOCmd.Mgmt_Bind_rsp: (
        STATUS,
        ("BindingTableEntries", t.uint8_t),
        ("StartIndex", t.uint8_t),
        ("BindingTableList", t.LVList[Binding, t.uint8_t]),
    ),
    # ... TODO optional stuff ...
    ZDOCmd.Mgmt_Leave_rsp: (STATUS,),
    ZDOCmd.Mgmt_Permit_Joining_rsp: (STATUS,),
    ZDOCmd.Mgmt_NWK_Update_rsp: (
        STATUS,
        ("ScannedChannels", t.Channels),
        ("TotalTransmissions", t.uint16_t),
        ("TransmissionFailures", t.uint16_t),
        ("EnergyValues", t.LVList[t.uint8_t, t.uint8_t]),
    ),
    # ... TODO optional stuff ...
}


# Rewrite to (name, param_names, param_types)
for command_id, schema in CLUSTERS.items():
    param_names = [p[0] for p in schema]
    param_types = [p[1] for p in schema]
    CLUSTERS[command_id] = (param_names, param_types)


class ZDOHeader:
    """Just a wrapper representing ZDO header, similar to ZCL header."""

    def __init__(self, command_id: t.uint16_t = 0x0000, tsn: t.uint8_t = 0) -> None:
        self._command_id = ZDOCmd(command_id)
        self._tsn = t.uint8_t(tsn)

    @property
    def command_id(self) -> ZDOCmd:
        """Return ZDO command."""
        return self._command_id

    @command_id.setter
    def command_id(self, value: t.uint16_t) -> None:
        """Command ID setter."""
        self._command_id = ZDOCmd(value)

    @property
    def is_reply(self) -> bool:
        """Return True if this is a reply."""
        return bool(self._command_id & 0x8000)

    @property
    def tsn(self) -> t.uint8_t:
        """Return transaction seq number."""
        return self._tsn

    @tsn.setter
    def tsn(self, value: t.uint8_t) -> None:
        """Set TSN."""
        self._tsn = t.uint8_t(value)

    @classmethod
    def deserialize(
        cls, command_id: t.uint16_t, data: bytes
    ) -> tuple[ZDOHeader, bytes]:
        """Deserialize data."""
        tsn, data = t.uint8_t.deserialize(data)
        return cls(command_id, tsn), data

    def serialize(self) -> bytes:
        """Serialize header."""

        return self.tsn.serialize()



import asyncio
import logging
from unittest.mock import Mock, patch

import pytest

from zigpy import datastructures


async def test_dynamic_bounded_semaphore_simple_locking():
    """Test simple, serial locking/unlocking."""
    sem = datastructures.PriorityDynamicBoundedSemaphore()

    assert "unlocked" not in repr(sem) and "locked" in repr(sem)

    assert sem.value == 0
    assert sem.max_value == 0
    assert sem.locked()

    # Invalid max value
    with pytest.raises(ValueError):
        sem.max_value = -1

    assert sem.value == 0
    assert sem.max_value == 0
    assert sem.locked()

    # Max value is now specified
    sem.max_value = 1
    assert not sem.locked()
    assert sem.max_value == 1
    assert sem.value == 1

    assert "unlocked" in repr(sem)

    # Semaphore can now be acquired
    async with sem:
        assert sem.value == 0
        assert sem.locked()

    assert not sem.locked()
    assert sem.max_value == 1
    assert sem.value == 1

    await sem.acquire()
    assert sem.value == 0
    assert sem.locked()
    sem.release()

    assert not sem.locked()
    assert sem.max_value == 1
    assert sem.value == 1

    with pytest.raises(ValueError):
        sem.release()


async def test_dynamic_bounded_semaphore_multiple_locking():
    """Test multiple locking/unlocking."""
    sem = datastructures.PriorityDynamicBoundedSemaphore(5)

    assert sem.value == 5
    assert not sem.locked()

    async with sem:
        assert sem.value == 4
        assert not sem.locked()

        async with sem, sem, sem:
            assert sem.value == 1
            assert not sem.locked()

            async with sem:
                assert sem.locked()
                assert sem.value == 0

            assert not sem.locked()
            assert sem.value == 1

        assert sem.value == 4
        assert not sem.locked()

    assert sem.value == 5
    assert not sem.locked()


async def test_dynamic_bounded_semaphore_hanging_bug():
    """Test semaphore hanging bug."""
    sem = datastructures.PriorityDynamicBoundedSemaphore(1)

    async def c1():
        async with sem:
            await asyncio.sleep(0)
        t2.cancel()

    async def c2():
        async with sem:
            pytest.fail("Should never get here")

    t1 = asyncio.create_task(c1())
    t2 = asyncio.create_task(c2())

    r1, r2 = await asyncio.gather(t1, t2, return_exceptions=True)
    assert r1 is None
    assert isinstance(r2, asyncio.CancelledError)

    assert not sem.locked()

    async with sem:
        assert True


def test_dynamic_bounded_semaphore_multiple_event_loops():
    """Test semaphore detecting multiple loops."""

    async def test_semaphore(sem):
        async with sem:
            await asyncio.sleep(0.1)

    async def make_semaphore():
        sem = datastructures.PriorityDynamicBoundedSemaphore(1)

        # The loop reference is lazily created so we need to actually lock the semaphore
        await asyncio.gather(test_semaphore(sem), test_semaphore(sem))

        return sem

    loop1 = asyncio.new_event_loop()
    sem = loop1.run_until_complete(make_semaphore())

    async def inner():
        await asyncio.gather(test_semaphore(sem), test_semaphore(sem))

    loop2 = asyncio.new_event_loop()

    with pytest.raises(RuntimeError):
        loop2.run_until_complete(inner())


async def test_dynamic_bounded_semaphore_runtime_limit_increase():
    """Test changing the max_value at runtime."""

    sem = datastructures.PriorityDynamicBoundedSemaphore(2)

    def set_limit(n):
        sem.max_value = n

    asyncio.get_running_loop().call_later(0.1, set_limit, 3)

    async with sem:
        # Play with the value, testing edge cases
        sem.max_value = 100
        assert sem.value == 99
        assert not sem.locked()

        sem.max_value = 2
        assert sem.value == 1
        assert not sem.locked()

        sem.max_value = 1
        assert sem.value == 0
        assert sem.locked()

        # Setting it to `0` seems undefined but we keep track of locks so it works
        sem.max_value = 0
        assert sem.value == -1
        assert sem.locked()

        sem.max_value = 2
        assert sem.value == 1
        assert not sem.locked()

        async with sem:
            assert sem.locked()
            assert sem.value == 0
            assert sem.max_value == 2

            async with sem:
                # We're now locked until the limit is increased
                pass

            assert not sem.locked()
            assert sem.value == 1
            assert sem.max_value == 3

        assert sem.value == 2
        assert sem.max_value == 3

    assert sem.value == 3
    assert sem.max_value == 3


async def test_dynamic_bounded_semaphore_errors():
    """Test semaphore handling errors and cancellation."""

    sem = datastructures.PriorityDynamicBoundedSemaphore(1)

    def set_limit(n):
        sem.max_value = n

    async def acquire():
        async with sem:
            await asyncio.sleep(60)

    # The first acquire call will succeed
    acquire1 = asyncio.create_task(acquire())

    # The remaining two will stall
    acquire2 = asyncio.create_task(acquire())
    acquire3 = asyncio.create_task(acquire())
    await asyncio.sleep(0.1)

    # Cancel the first one, which holds the lock
    acquire1.cancel()

    # But also cancel the second one, which was waiting
    acquire2.cancel()
    with pytest.raises(asyncio.CancelledError):
        await acquire1

    with pytest.raises(asyncio.CancelledError):
        await acquire2

    await asyncio.sleep(0.1)

    # The third one will have succeeded
    assert sem.locked()
    assert sem.value == 0
    assert sem.max_value == 1

    acquire3.cancel()
    with pytest.raises(asyncio.CancelledError):
        await acquire3

    assert not sem.locked()
    assert sem.value == 1
    assert sem.max_value == 1


async def test_dynamic_bounded_semaphore_cancellation():
    """Test semaphore handling errors and cancellation."""

    sem = datastructures.PriorityDynamicBoundedSemaphore(2)

    async def acquire():
        async with sem:
            await asyncio.sleep(0.2)

    tasks = []

    # First two lock up the semaphore but succeed
    tasks.append(asyncio.create_task(acquire()))
    tasks.append(asyncio.create_task(acquire()))

    # Next two get in line, will be cancelled
    tasks.append(asyncio.create_task(acquire()))
    tasks.append(asyncio.create_task(acquire()))

    await asyncio.sleep(0)
    exc = RuntimeError("Uh oh :(")
    sem.cancel_waiting(exc)

    # Last one makes it through
    tasks.append(asyncio.create_task(acquire()))

    assert (await asyncio.gather(*tasks, return_exceptions=True)) == [
        None,
        None,
        exc,
        exc,
        None,
    ]

    assert not sem.locked()


async def test_dynamic_bounded_semaphore_cancel_waiting_then_task_cancelled() -> None:
    """Test that a cancelled waiter does not corrupt the active-request bookkeeping."""
    sem = datastructures.PriorityDynamicBoundedSemaphore(1)

    async def acquire():
        async with sem:
            await asyncio.sleep(60)

    # Hold the only slot so the next acquire must wait
    async with sem:
        assert sem.value == 0

        task = asyncio.create_task(acquire())
        await asyncio.sleep(0)  # Let the task start waiting
        assert len(sem._waiters) == 1

        # `cancel_waiting` resolves the future with an exception and the task is
        # then cancelled: cancellation wins, so the acquire raises `CancelledError`
        # even though no slot was ever granted.
        sem.cancel_waiting(RuntimeError("rejoin"))
        task.cancel()

        with pytest.raises(asyncio.CancelledError):
            await task

        # The waiter never acquired a slot, so the active count is unchanged
        assert sem.value == 0

    # Releasing the outer slot must not trip the bookkeeping assertion
    assert sem.value == 1


async def test_priority_lock():
    """Test priority lock."""

    lock = datastructures.PriorityLock()

    with pytest.raises(ValueError):
        lock.max_value = 2

    assert lock.max_value == 1

    # Default priority of 0
    async with lock:
        pass

    # Overridden priority of 100
    async with lock(priority=100):
        pass

    run_order = []

    async def test_priority(priority: int, item: str):
        assert lock.locked()

        async with lock(priority=priority):
            run_order.append(item)

    # Lock first
    async with lock:
        assert lock.locked()

        names = {
            "1: first": 1,
            "5: first": 5,
            "1: second": 1,
            "1: third": 1,
            "5: second": 5,
            "-5: only": -5,
            "1: fourth": 1,
            "2: only": 2,
        }

        tasks = {
            name: asyncio.create_task(test_priority(priority + 0, name + ""))
            for name, priority in names.items()
        }

        await asyncio.sleep(0)
        tasks["1: second"].cancel()
        await asyncio.sleep(0)

    await asyncio.gather(*tasks.values(), return_exceptions=True)

    assert run_order == [
        "5: first",
        "5: second",
        "2: only",
        "1: first",
        # "1: second",
        "1: third",
        "1: fourth",
        "-5: only",
    ]


async def test_reschedulable_timeout():
    callback = Mock()
    timeout = datastructures.ReschedulableTimeout(callback)

    timeout.reschedule(0.1)
    assert len(callback.mock_calls) == 0
    await asyncio.sleep(0.09)
    assert len(callback.mock_calls) == 0
    await asyncio.sleep(0.02)
    assert len(callback.mock_calls) == 1


async def test_reschedulable_timeout_reschedule():
    callback = Mock()
    timeout = datastructures.ReschedulableTimeout(callback)

    timeout.reschedule(0.1)
    timeout.reschedule(0.2)
    await asyncio.sleep(0.19)
    assert len(callback.mock_calls) == 0
    await asyncio.sleep(0.02)
    assert len(callback.mock_calls) == 1


async def test_reschedulable_timeout_cancel():
    callback = Mock()
    timeout = datastructures.ReschedulableTimeout(callback)

    timeout.reschedule(0.1)
    assert len(callback.mock_calls) == 0
    await asyncio.sleep(0.09)
    timeout.cancel()
    await asyncio.sleep(0.02)
    assert len(callback.mock_calls) == 0


async def test_debouncer():
    """Test debouncer."""

    debouncer = datastructures.Debouncer()
    debouncer.clean()
    assert repr(debouncer) == "<Debouncer [tracked:0]>"

    obj1 = object()
    assert not debouncer.is_filtered(obj1)
    assert not debouncer.filter(obj1, expire_in=0.1)
    assert debouncer.is_filtered(obj1)

    assert debouncer.filter(obj1, expire_in=1)
    assert debouncer.filter(obj1, expire_in=0.1)
    assert debouncer.filter(obj1, expire_in=1)
    assert debouncer.is_filtered(obj1)
    assert repr(debouncer) == "<Debouncer [tracked:1]>"

    obj2 = object()
    assert not debouncer.is_filtered(obj2)
    assert not debouncer.filter(obj2, expire_in=0.2)
    assert debouncer.filter(obj1, expire_in=1)
    assert debouncer.filter(obj2, expire_in=1)
    assert debouncer.filter(obj1, expire_in=1)
    assert debouncer.filter(obj2, expire_in=1)

    assert debouncer.is_filtered(obj1)
    assert debouncer.is_filtered(obj2)
    assert repr(debouncer) == "<Debouncer [tracked:2]>"

    await asyncio.sleep(0.1)
    assert not debouncer.is_filtered(obj1)
    assert debouncer.is_filtered(obj2)
    assert repr(debouncer) == "<Debouncer [tracked:1]>"

    await asyncio.sleep(0.1)
    assert not debouncer.is_filtered(obj1)
    assert not debouncer.is_filtered(obj2)
    assert repr(debouncer) == "<Debouncer [tracked:0]>"


async def test_debouncer_low_resolution_clock():
    """Test debouncer with a low resolution clock."""

    loop = asyncio.get_running_loop()
    now = loop.time()

    # Make sure we can debounce on a low resolution clock
    with patch.object(loop, "time", return_value=now):
        debouncer = datastructures.Debouncer()

        obj1 = object()
        debouncer.filter(obj1, expire_in=0.1)
        assert debouncer.is_filtered(obj1)

        obj2 = object()
        debouncer.filter(obj2, expire_in=0.1)
        assert debouncer.is_filtered(obj2)

        # The two objects cannot be compared
        with pytest.raises(TypeError):
            obj1 < obj2  # noqa: B015


async def test_debouncer_cleaning_bug():
    """Test debouncer bug when using heapq improperly."""
    debouncer = datastructures.Debouncer()

    obj1 = object()
    obj2 = object()
    obj3 = object()

    # Filter obj1 with an expiration of 0.3 seconds
    debouncer.filter(obj1, expire_in=0.3)

    # Slight delay to ensure different expiration times
    await asyncio.sleep(0.05)

    # Filter obj2 with an expiration of 0.1 seconds
    debouncer.filter(obj2, expire_in=0.1)

    # Another slight delay
    await asyncio.sleep(0.05)

    # Filter obj3 with an expiration of 0.2 seconds
    debouncer.filter(obj3, expire_in=0.2)

    assert debouncer.is_filtered(obj1)
    assert debouncer.is_filtered(obj2)
    assert debouncer.is_filtered(obj3)

    # Wait until after obj2 should have expired
    await asyncio.sleep(0.11)  # Total elapsed time ~0.21 seconds from start

    # Clean up expired items
    debouncer.clean()

    # obj2 should have expired, but due to the bug, it might still be filtered
    assert not debouncer.is_filtered(obj2)

    # obj1 and obj3 should still be filtered
    assert debouncer.is_filtered(obj1)
    assert debouncer.is_filtered(obj3)

    # Wait until after obj1 and obj3 should have expired
    await asyncio.sleep(0.1)  # Total elapsed time ~0.31 seconds from start

    # Clean up expired items
    debouncer.clean()

    # Now all objects should have expired
    assert not debouncer.is_filtered(obj1)
    assert not debouncer.is_filtered(obj3)

    # The queue should be empty
    assert len(debouncer._queue) == 0


async def test_request_limiter_simple_tier():
    """Test the limiter behaving like a simple semaphore with one tier."""
    limiter = datastructures.RequestLimiter(2, {0: 1})
    results = []
    order = []

    async def worker(uid):
        nonlocal order
        order.append(f"trying-{uid}")
        async with limiter(priority=0):
            order.append(f"acquired-{uid}")
            results.append(f"start-{uid}")
            await asyncio.sleep(0.05)
            results.append(f"end-{uid}")

    tasks = [asyncio.create_task(worker(i)) for i in range(3)]

    # The first two should start immediately
    await asyncio.sleep(0)
    assert limiter.active_requests == 2
    assert limiter.waiting_requests == 1
    assert "active=2, waiting=1" in repr(limiter)

    # Wait for all to complete
    await asyncio.gather(*tasks)

    assert limiter.active_requests == 0
    assert limiter.waiting_requests == 0
    # The exact end order depends on scheduling, but the start order is what matters.
    assert results[:2] == ["start-0", "start-1"]
    assert results[2:4] == ["end-0", "end-1"]
    assert results[4] == "start-2"
    assert results[5] == "end-2"


async def test_request_limiter_cascading_priority():
    """Test the core cascading priority logic."""
    # 2 slots for low priority, 1 extra for high priority (total 3)
    limiter = datastructures.RequestLimiter(3, {0: 2 / 3, 10: 1 / 3})
    events = []

    async def worker(uid, priority, delay):
        nonlocal events
        async with limiter(priority=priority):
            events.append(f"start-{uid}(p={priority})")
            await asyncio.sleep(delay)
            events.append(f"end-{uid}(p={priority})")

    # These two should take the low-priority slots
    t1 = asyncio.create_task(worker(1, 0, 0.1))
    t2 = asyncio.create_task(worker(2, 0, 0.1))

    # This one should have to wait
    t3 = asyncio.create_task(worker(3, 0, 0.1))

    await asyncio.sleep(0)
    assert limiter.active_requests == 2
    assert limiter.waiting_requests == 1
    assert events == ["start-1(p=0)", "start-2(p=0)"]

    # This high-priority one should be able to acquire the reserved slot
    t4 = asyncio.create_task(worker(4, 10, 0.1))
    await asyncio.sleep(0)
    assert limiter.active_requests == 3
    assert limiter.waiting_requests == 1
    assert "start-4(p=10)" in events

    # No more slots available for anyone
    t5 = asyncio.create_task(worker(5, 10, 0.1))
    await asyncio.sleep(0)
    assert limiter.active_requests == 3
    assert limiter.waiting_requests == 2

    await asyncio.gather(t1, t2, t3, t4, t5)

    assert limiter.active_requests == 0
    assert limiter.waiting_requests == 0


async def test_request_limiter_priority_queueing():
    """Test that waiters are woken up according to their priority."""
    limiter = datastructures.RequestLimiter(1, {0: 1})
    events = []

    async def worker(uid, priority):
        nonlocal events
        async with limiter(priority=priority):
            events.append(uid)
            await asyncio.sleep(0.1)  # Hold the lock long enough for others to queue

    # First task takes the only slot
    t1 = asyncio.create_task(worker(1, 0))
    await asyncio.sleep(0)  # Ensure t1 acquires the lock
    assert limiter.active_requests == 1

    # Queue up waiters with different priorities
    t2 = asyncio.create_task(worker(2, 5))  # medium
    t3 = asyncio.create_task(worker(3, 0))  # low
    t4 = asyncio.create_task(worker(4, 10))  # high

    await asyncio.sleep(0)  # Ensure all waiters have tried to acquire
    assert limiter.waiting_requests == 3

    # Wait for all tasks to complete
    await asyncio.gather(t1, t2, t3, t4)

    # The waiters should have run in priority order, not submission order
    assert events == [1, 4, 2, 3]


async def test_request_limiter_cancellation():
    """Test cancelling tasks that are waiting for or holding a slot."""
    limiter = datastructures.RequestLimiter(1, {0: 1})

    async def holder():
        async with limiter(priority=0):
            await asyncio.sleep(60)

    async def waiter():
        async with limiter(priority=0):
            pytest.fail("Waiter should not acquire the slot")

    # t1 acquires the only slot
    t1 = asyncio.create_task(holder())
    await asyncio.sleep(0)
    assert limiter.active_requests == 1

    # t2 waits in line
    t2 = asyncio.create_task(waiter())
    await asyncio.sleep(0)
    assert limiter.waiting_requests == 1

    # Cancel the waiting task
    t2.cancel()
    with pytest.raises(asyncio.CancelledError):
        await t2

    assert limiter.waiting_requests == 0
    assert limiter.active_requests == 1

    # Now, cancel the holder task
    t1.cancel()
    with pytest.raises(asyncio.CancelledError):
        await t1

    assert limiter.active_requests == 0
    assert limiter.waiting_requests == 0

    # The limiter should be usable again
    async with limiter(priority=0):
        assert limiter.active_requests == 1


async def test_request_limiter_invalid_priority():
    """Test that using a priority with no allocated tier raises an error."""
    limiter = datastructures.RequestLimiter(1, {10: 1})

    with pytest.raises(
        ValueError,
        match="Priority 5 is lower than the lowest known priority 10 and has no allocated capacity",
    ):
        async with limiter(priority=5):
            pytest.fail("This code should not be reached")


async def test_request_limiter_highest_priority():
    """Test that the highest priority task is always allowed to run."""
    limiter = datastructures.RequestLimiter(1, {100: 1})

    async with limiter(priority=100 + 1):
        assert limiter.active_requests == 1

        assert limiter.locked(priority=100)


async def test_request_limiter_max_concurrency_property():
    """Test the max_concurrency property getter."""
    limiter = datastructures.RequestLimiter(5, {1: 1.0})
    assert limiter.max_concurrency == 5


async def test_request_limiter_priority_higher_than_known():
    """Test priority higher than the highest known tier."""
    limiter = datastructures.RequestLimiter(2, {5: 0.5, 10: 0.5})

    # Priority 15 is higher than highest known (10), should use highest tier
    async with limiter(priority=15):
        assert limiter.active_requests == 1


async def test_request_limiter_cancel_waiting():
    """Test cancelling all waiting tasks."""
    limiter = datastructures.RequestLimiter(1, {1: 1.0})

    # Acquire the limiter first
    async with limiter(priority=1):
        # Create waiting tasks
        async def waiter():
            await limiter._acquire(priority=1)

        task1 = asyncio.create_task(waiter())
        task2 = asyncio.create_task(waiter())
        await asyncio.sleep(0)  # Let tasks start waiting

        assert limiter.waiting_requests == 2

        # Cancel all waiting tasks
        exc = RuntimeError("Test cancellation")
        limiter.cancel_waiting(exc)

        # Tasks should be cancelled with the provided exception
        with pytest.raises(RuntimeError, match="Test cancellation"):
            await task1
        with pytest.raises(RuntimeError, match="Test cancellation"):
            await task2


async def test_request_limiter_cancel_waiting_then_task_cancelled() -> None:
    """Test that a cancelled waiter does not corrupt the active-request bookkeeping."""
    limiter = datastructures.RequestLimiter(1, {1: 1.0})

    async def acquire():
        async with limiter(priority=1):
            await asyncio.sleep(60)

    # Hold the only slot so the next acquire must wait
    async with limiter(priority=1):
        assert limiter.active_requests == 1

        task = asyncio.create_task(acquire())
        await asyncio.sleep(0)  # Let the task start waiting
        assert limiter.waiting_requests == 1

        # `cancel_waiting` resolves the future with an exception and the task is
        # then cancelled: cancellation wins, so the acquire raises `CancelledError`
        # even though no slot was ever granted.
        limiter.cancel_waiting(RuntimeError("rejoin"))
        task.cancel()

        with pytest.raises(asyncio.CancelledError):
            await task

        # The waiter never acquired a slot, so the active count is unchanged
        assert limiter.active_requests == 1

    # Releasing the outer slot must not trip the bookkeeping assertion
    assert limiter.active_requests == 0


async def test_request_limiter_backwards_compatibility() -> None:
    """Test `max_value` property proxying `max_concurrency`."""
    limiter = datastructures.RequestLimiter(5, {1: 1.0})

    with pytest.deprecated_call():
        assert limiter.max_value == 5

    with pytest.deprecated_call():
        limiter.max_value = 10

    assert limiter.max_concurrency == 10

    with pytest.deprecated_call():
        assert limiter.max_value == 10


async def test_request_limiter_adjust_max_concurrency(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test adjusting max concurrency at runtime snaps to a reasonable value."""
    limiter = datastructures.RequestLimiter(8, {1: 1 / 3, 2: 1 / 2, 3: 1 / 6})

    assert limiter.max_concurrency == 12

    with caplog.at_level(logging.WARNING):
        limiter.max_concurrency = 13

    # It'll be bumped to 18, which is divisible by 3, 2, and 6
    assert limiter.max_concurrency == 18



"""Test configuration."""

import pathlib
import warnings

import pytest
import voluptuous as vol

import zigpy.config
import zigpy.config.validators


@pytest.mark.parametrize(
    ("value", "result"),
    [
        (False, False),
        (True, True),
        ("1", True),
        ("yes", True),
        ("YeS", True),
        ("on", True),
        ("oN", True),
        ("enable", True),
        ("enablE", True),
        (0, False),
        ("no", False),
        ("nO", False),
        ("off", False),
        ("ofF", False),
        ("disable", False),
        ("disablE", False),
    ],
)
def test_config_validation_bool(value, result):
    """Test boolean config validation."""
    assert zigpy.config.validators.cv_boolean(value) is result

    schema = vol.Schema({vol.Required("value"): zigpy.config.validators.cv_boolean})
    validated = schema({"value": value})
    assert validated["value"] is result


@pytest.mark.parametrize("value", ["invalid", "not a bool", "something"])
def test_config_validation_bool_invalid(value):
    """Test boolean config validation."""
    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_boolean(value)


def test_config_validation_key_not_16_list():
    """Validate key fails."""
    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0x00])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0x00 for i in range(15)])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0x00 for i in range(17)])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key(None)

    zigpy.config.validators.cv_key([0x00 for i in range(16)])


def test_config_validation_key_not_a_byte():
    """Validate key fails."""

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([-1 for i in range(16)])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([256 for i in range(16)])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0] * 15 + [256])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0] * 15 + [-1])

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_key([0] * 15 + ["x1"])

    zigpy.config.validators.cv_key([0xFF for i in range(16)])


def test_config_validation_key_success():
    """Validate key success."""

    key = zigpy.config.validators.cv_key(zigpy.config.CONF_NWK_TC_LINK_KEY_DEFAULT)
    assert key.serialize() == b"ZigBeeAlliance09"


@pytest.mark.parametrize(
    ("value", "result"),
    [
        (0x1234, 0x1234),
        ("0x1234", 0x1234),
        (1234, 1234),
        ("1234", 1234),
        ("001234", 1234),
        ("0e1234", vol.Invalid),
        ("1234abcd", vol.Invalid),
        ("0xabGG", vol.Invalid),
        (None, vol.Invalid),
    ],
)
def test_config_validation_hex_number(value, result):
    """Test hex number config validation."""

    if isinstance(result, int):
        assert zigpy.config.validators.cv_hex(value) == result
    else:
        with pytest.raises(vol.Invalid):
            zigpy.config.validators.cv_hex(value)


@pytest.mark.parametrize(
    ("value", "result"),
    [
        (1, vol.Invalid),
        (11, 11),
        (0x11, 17),
        ("26", 26),
        (27, vol.Invalid),
        ("27", vol.Invalid),
    ],
)
def test_schema_network_channel(value, result):
    """Test network schema for channel."""

    config = {zigpy.config.CONF_NWK_CHANNEL: value}

    if isinstance(result, int):
        config = zigpy.config.SCHEMA_NETWORK(config)
        assert config[zigpy.config.CONF_NWK_CHANNEL] == result
    else:
        with pytest.raises(vol.Invalid):
            zigpy.config.SCHEMA_NETWORK(config)


def test_schema_network_pan_id():
    """Test Extended Pan-id."""
    config = zigpy.config.SCHEMA_NETWORK({})
    assert (
        config[zigpy.config.CONF_NWK_EXTENDED_PAN_ID]
        == zigpy.config.CONF_NWK_EXTENDED_PAN_ID_DEFAULT
    )

    config = zigpy.config.SCHEMA_NETWORK(
        {zigpy.config.CONF_NWK_EXTENDED_PAN_ID: "00:11:22:33:44:55:66:77"}
    )
    assert (
        config[zigpy.config.CONF_NWK_EXTENDED_PAN_ID].serialize()
        == b"\x77\x66\x55\x44\x33\x22\x11\x00"
    )


def test_schema_network_short_pan_id():
    """Test Pan-id."""
    config = zigpy.config.SCHEMA_NETWORK({})
    assert config[zigpy.config.CONF_NWK_PAN_ID] is None

    config = zigpy.config.SCHEMA_NETWORK({zigpy.config.CONF_NWK_PAN_ID: 0x1234})
    assert config[zigpy.config.CONF_NWK_PAN_ID].serialize() == b"\x34\x12"


def test_deprecated():
    """Test key deprecation."""

    schema = vol.Schema(
        {
            vol.Optional("value"): vol.All(
                zigpy.config.validators.cv_hex,
                zigpy.config.validators.cv_deprecated("Test message"),
            )
        }
    )

    with pytest.warns(DeprecationWarning, match="Test message"):
        assert schema({"value": 123}) == {"value": 123}

    # No warnings are raised
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert schema({}) == {}


def test_cv_json_file(tmp_path: pathlib.Path) -> None:
    """Test `cv_json_file` validator."""

    path = tmp_path / "file.json"

    # Does not exist
    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_json_file(str(path))

    # Not a file
    path.mkdir()

    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_json_file(str(path))

    path.rmdir()

    # File exists
    path.write_text("{}")
    assert zigpy.config.validators.cv_json_file(str(path)) == path


def test_cv_folder(tmp_path: pathlib.Path) -> None:
    """Test `cv_folder` validator."""

    folder_path = tmp_path / "folder"
    file_path = tmp_path / "not_folder"

    # Does not exist
    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_folder(str(folder_path))

    # Not a folder
    file_path.write_text("")
    with pytest.raises(vol.Invalid):
        zigpy.config.validators.cv_folder(str(file_path))

    # Folder exists
    folder_path.mkdir()
    assert zigpy.config.validators.cv_folder(str(folder_path)) == folder_path


@pytest.mark.parametrize(
    ("value", "limit", "should_warn"),
    [
        (5, 10, False),
        (10, 10, False),
        (11, 10, True),
        (20, 10, True),
        (-5, 0, False),
        (1, 0, True),
    ],
)
def test_cv_warn_if_greater(value, limit, should_warn):
    """Test `cv_warn_if_greater` validator."""
    message = "Test warning message"
    validator = zigpy.config.validators.cv_warn_if_greater(limit, message)

    if should_warn:
        with pytest.warns(UserWarning, match=message):
            assert validator(value) == value
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            assert validator(value) == value



"""Regenerate per-module mypy disable_error_code overrides in pyproject.toml.

Usage:
    python -m tools.regenerate_mypy_ignores
"""

from collections import defaultdict
from pathlib import Path
import re

from mypy import api as mypy_api

START_MARKER = "# Autogenerated mypy overrides: start"
END_MARKER = "# Autogenerated mypy overrides: end"


def run_mypy() -> list[str]:
    """Run mypy with no error codes disabled and capture output."""
    stdout, stderr, _ = mypy_api.run(
        [
            "zigpy",
            "--no-error-summary",
            "--config-file=/dev/null",
            "--ignore-missing-imports",
        ]
    )
    return stdout.splitlines() + stderr.splitlines()


def parse_errors(lines: list[str]) -> dict[str, dict[str, int]]:
    """Parse mypy output into {module: {error_code: count}}."""
    errors_by_module: dict[str, dict[str, int]] = defaultdict(lambda: defaultdict(int))

    for line in lines:
        match = re.match(r"^([^:]+):.* \[([a-z-]+)\]$", line.strip())
        if match:
            filepath, error_code = match.groups()
            module = filepath.replace("/", ".").replace(".py", "")
            module = module.removesuffix(".__init__")
            errors_by_module[module][error_code] += 1

    return dict(errors_by_module)


def format_override(module: str, codes: dict[str, int]) -> str:
    """Format a single [[tool.mypy.overrides]] section."""
    lines = [
        "[[tool.mypy.overrides]]",
        f'module = "{module}"',
        "disable_error_code = [",
    ]
    for code in sorted(codes.keys()):
        lines.append(f'    "{code}",  # {codes[code]}')  # noqa: PERF401
    lines.append("]")
    return "\n".join(lines)


def generate_overrides(errors_by_module: dict[str, dict[str, int]]) -> str:
    """Generate all override sections."""
    # Aggregate zigpy.zcl.clusters.* into one entry
    clusters_errors: dict[str, int] = defaultdict(int)
    non_clusters: dict[str, dict[str, int]] = {}

    for module, codes in errors_by_module.items():
        if module.startswith("zigpy.zcl.clusters.") or module == "zigpy.zcl.clusters":
            for code, count in codes.items():
                clusters_errors[code] += count
        else:
            non_clusters[module] = codes

    # Sort by (total error count, module name) so easiest fixes are at the top
    sorted_modules = sorted(
        non_clusters.items(), key=lambda x: (sum(x[1].values()), x[0])
    )

    sections = [format_override(module, codes) for module, codes in sorted_modules]

    # Add aggregated clusters entry at the end
    if clusters_errors:
        sections.append(format_override("zigpy.zcl.clusters.*", dict(clusters_errors)))

    return "\n\n".join(sections)


def main():
    pyproject_path = Path("pyproject.toml")
    content = pyproject_path.read_text()

    lines = run_mypy()
    errors_by_module = parse_errors(lines)
    overrides_content = generate_overrides(errors_by_module)

    # Replace content between markers
    pattern = re.compile(
        rf"({re.escape(START_MARKER)})\n.*?\n({re.escape(END_MARKER)})",
        re.DOTALL,
    )
    new_content = pattern.sub(rf"\1\n{overrides_content}\n\2", content)

    if new_content != content:
        pyproject_path.write_text(new_content)


if __name__ == "__main__":
    main()



from __future__ import annotations

import inspect

from .. import Cluster
from . import (
    closures,
    general,
    general_const as general_const,  # noqa: PLC0414
    greenpower,
    homeautomation,
    hvac,
    lighting,
    lightlink,
    manufacturer_specific,
    measurement,
    protocol,
    security,
    smartenergy,
    wwah,
)

CLUSTERS_BY_ID: dict[int, Cluster] = {}
CLUSTERS_BY_NAME: dict[str, Cluster] = {}

for cls in (
    closures,
    general,
    greenpower,
    homeautomation,
    hvac,
    lighting,
    lightlink,
    manufacturer_specific,
    measurement,
    protocol,
    security,
    smartenergy,
    wwah,
):
    for name in dir(cls):
        obj = getattr(cls, name)

        # Object must be a concrete Cluster subclass
        if (
            not inspect.isclass(obj)
            or not issubclass(obj, Cluster)
            or obj.cluster_id is None
        ):
            continue

        assert CLUSTERS_BY_ID.get(obj.cluster_id, obj) is obj
        assert CLUSTERS_BY_NAME.get(obj.ep_attribute, obj) is obj

        CLUSTERS_BY_ID[obj.cluster_id] = obj
        CLUSTERS_BY_NAME[obj.ep_attribute] = obj



"""Constants Related to General Clusters"""

from __future__ import annotations

import zigpy.types as t


class AnalogInputType(t.enum8):
    Temp_Degrees_C = 0x00
    Relative_Humidity_Percent = 0x01
    Pressure_Pascal = 0x02
    Flow_Liters_Per_Sec = 0x03
    Percentage = 0x04
    Parts_Per_Million = 0x05
    Rotational_Speed_RPM = 0x06
    Current_Amps = 0x07
    Frequency_Hz = 0x08
    Power_Watts = 0x09
    Power_Kilo_Watts = 0x0A
    Energy_Kilo_Watt_Hours = 0x0B
    Count = 0x0C
    Enthalpy_KJoules_Per_Kg = 0x0D
    Time_Seconds = 0x0E


class TempDegreesC(t.enum16):
    Two_Pipe_Entering_Water_Temperature = 0x0000
    Two_Pipe_Leaving_Water_Temperature = 0x0001
    Boiler_Entering_Temperature = 0x0002
    Boiler_Leaving_Temperature = 0x0003
    Chiller_Chilled_Water_Entering_Temp = 0x0004
    Chiller_Chilled_Water_Leaving_Temp = 0x0005
    Chiller_Condenser_Water_Entering_Temp = 0x0006
    Chiller_Condenser_Water_Leaving_Temp = 0x0007
    Cold_Deck_Temperature = 0x0008
    Cooling_Coil_Discharge_Temperature = 0x0009
    Cooling_Entering_Water_Temperature = 0x000A
    Cooling_Leaving_Water_Temperature = 0x000B
    Condenser_Water_Return_Temperature = 0x000C
    Condenser_Water_Supply_Temperature = 0x000D
    Decouple_Loop_Temperature = 0x000E
    Building_Load = 0x000F
    Decouple_Loop_Temperature_2 = 0x0010
    Dew_Point_Temperature = 0x0011
    Discharge_Air_Temperature = 0x0012
    Discharge_Temperature = 0x0013
    Exhaust_Air_Temperature_After_Heat_Recovery = 0x0014
    Exhaust_Air_Temperature = 0x0015
    Glycol_Temperature = 0x0016
    Heat_Recovery_Air_Temperature = 0x0017
    Hot_Deck_Temperature = 0x0018
    Heat_Exchanger_Bypass_Temp = 0x0019
    Heat_Exchanger_Entering_Temp = 0x001A
    Heat_Exchanger_Leaving_Temp = 0x001B
    Mechanical_Room_Temperature = 0x001C
    Mixed_Air_Temperature = 0x001D
    Mixed_Air_Temperature_2 = 0x001E
    Outdoor_Air_Dewpoint_Temp = 0x001F
    Outdoor_Air_Temperature = 0x0020
    Preheat_Air_Temperature = 0x0021
    Preheat_Entering_Water_Temperature = 0x0022
    Preheat_Leaving_Water_Temperature = 0x0023
    Primary_Chilled_Water_Return_Temp = 0x0024
    Primary_Chilled_Water_Supply_Temp = 0x0025
    Primary_Hot_Water_Return_Temp = 0x0026
    Primary_Hot_Water_Supply_Temp = 0x0027
    Reheat_Coil_Discharge_Temperature = 0x0028
    Reheat_Entering_Water_Temperature = 0x0029
    Reheat_Leaving_Water_Temperature = 0x002A
    Return_Air_Temperature = 0x002B
    Secondary_Chilled_Water_Return_Temp = 0x002C
    Secondary_Chilled_Water_Supply_Temp = 0x002D
    Secondary_HW_Return_Temp = 0x002E
    Secondary_HW_Supply_Temp = 0x002F
    Sideloop_Reset_Temperature = 0x0030
    Sideloop_Temperature_Setpoint = 0x0031
    Sideloop_Temperature = 0x0032
    Source_Temperature = 0x0033
    Supply_Air_Temperature = 0x0034
    Supply_Low_Limit_Temperature = 0x0035
    Tower_Basin_Temp = 0x0036
    Two_Pipe_Leaving_Water_Temp = 0x0037
    Reserved = 0x0038
    Zone_Dewpoint_Temperature = 0x0039
    Zone_Sensor_Setpoint = 0x003A
    Zone_Sensor_Setpoint_Offset = 0x003B
    Zone_Temperature = 0x003C
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class RelativeHumidityPercent(t.enum16):
    Discharge_Humidity = 0x0000
    Exhaust_Humidity = 0x0001
    Hot_Deck_Humidity = 0x0002
    Mixed_Air_Humidity = 0x0003
    Outdoor_Air_Humidity = 0x0004
    Return_Humidity = 0x0005
    Sideloop_Humidity = 0x0006
    Space_Humidity = 0x0007
    Zone_Humidity = 0x0008
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class PressurePascal(t.enum16):
    Boiler_Pump_Differential_Pressure = 0x0000
    Building_Static_Pressure = 0x0001
    Cold_Deck_Differential_Pressure_Sensor = 0x0002
    Chilled_Water_Building_Differential_Pressure = 0x0003
    Cold_Deck_Differential_Pressure = 0x0004
    Cold_Deck_Static_Pressure = 0x0005
    Condenser_Water_Pump_Differential_Pressure = 0x0006
    Discharge_Differential_Pressure = 0x0007
    Discharge_Static_Pressure_1 = 0x0008
    Discharge_Static_Pressure_2 = 0x0009
    Exhaust_Air_Differential_Pressure = 0x000A
    Exhaust_Air_Static_Pressure = 0x000B
    Exhaust_Differential_Pressure = 0x000C
    Exhaust_Differential_Pressure_2 = 0x000D
    Hot_Deck_Differential_Pressure = 0x000E
    Hot_Deck_Differential_Pressure_2 = 0x000F
    Hot_Deck_Static_Pressure = 0x0010
    Hot_Water_Bldg_Diff_Pressure = 0x0011
    Heat_Exchanger_Steam_Pressure = 0x0012
    Minimum_Outdoor_Air_Differential_Pressure = 0x0013
    Outdoor_Air_Differential_Pressure = 0x0014
    Primary_Chilled_Water_Pump_Differential_Pressure = 0x0015
    Primary_Hot_Water_Pump_Differential_Pressure = 0x0016
    Relief_Differential_Pressure = 0x0017
    Return_Air_Static_Pressure = 0x0018
    Return_Differential_Pressure = 0x0019
    Secondary_Chilled_Water_Pump_Differential_Pressure = 0x001A
    Secondary_Hot_Water_Pump_Differential_Pressure = 0x001B
    Sideloop_Pressure = 0x001C
    Steam_Pressure = 0x001D
    Supply_Differential_Pressure_Sensor = 0x001E
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class FlowLitersPerSec(t.enum16):
    Chilled_Water_Flow = 0x0000
    Chiller_Chilled_Water_Flow = 0x0001
    Chiller_Condenser_Water_Flow = 0x0002
    Cold_Deck_Flow = 0x0003
    Decouple_Loop_Flow = 0x0004
    Discharge_Flow = 0x0005
    Exhaust_Fan_Flow = 0x0006
    Exhaust_Flow = 0x0007
    Fan_Flow = 0x0008
    Hot_Deck_Flow = 0x0009
    Hot_Water_Flow = 0x000A
    Minimum_Outdoor_Air_Fan_Flow = 0x000B
    Minimum_Outdoor_Air_Flow = 0x000C
    Outdoor_Air_Flow = 0x000D
    Primary_Chilled_Water_Flow = 0x000E
    Relief_Fan_Flow = 0x000F
    Relief_Flow = 0x0010
    Return_Fan_Flow = 0x0011
    Return_Flow = 0x0012
    Secondary_Chilled_Water_Flow = 0x0013
    Supply_Fan_Flow = 0x0014
    Tower_Fan_Flow = 0x0015
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class Percentage(t.enum16):
    Chiller_Percent_Full_Load_Amperage = 0x000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class PartsPerMillion(t.enum16):
    Return_Carbon_Dioxide = 0x0000
    Zone_Carbon_Dioxide = 0x0001
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class RotationalSpeedRPM(t.enum16):
    Exhaust_Fan_Remote_Speed = 0x0000
    Heat_Recovery_Wheel_Remote_Speed = 0x0001
    Min_Outdoor_Air_Fan_Remote_Speed = 0x0002
    Relief_Fan_Remote_Speed = 0x0003
    Return_Fan_Remote_Speed = 0x0004
    Supply_Fan_Remote_Speed = 0x0005
    Variable_Speed_Drive_Motor_Speed = 0x0006
    Variable_Speed_Drive_Speed_Setpoint = 0x0007
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class CurrentAmps(t.enum16):
    Chiller_Amps = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class FrequencyHz(t.enum16):
    Variable_Speed_Drive_Output_Frequency = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class PowerWatts(t.enum16):
    Power_Consumption = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class PowerKiloWatts(t.enum16):
    Absolute_Power = 0x0000
    Power_Consumption = 0x0001
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class EnergyKiloWattHours(t.enum16):
    Variable_Speed_Drive_Kilowatt_Hours = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class Count(t.enum16):
    Count = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class EnthalpyKJoulesPerKg(t.enum16):
    Outdoor_Air_Enthalpy = 0x0000
    Return_Air_Enthalpy = 0x0001
    Space_Enthalpy = 0x0002
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


class TimeSeconds(t.enum16):
    Relative_Time = 0x0000
    # 0x0200 through 0xFFFE are vendor defined
    Other = 0xFFFF


ANALOG_INPUT_TYPES = {
    AnalogInputType.Temp_Degrees_C: TempDegreesC,
    AnalogInputType.Relative_Humidity_Percent: RelativeHumidityPercent,
    AnalogInputType.Pressure_Pascal: PressurePascal,
    AnalogInputType.Flow_Liters_Per_Sec: FlowLitersPerSec,
    AnalogInputType.Percentage: Percentage,
    AnalogInputType.Parts_Per_Million: PartsPerMillion,
    AnalogInputType.Rotational_Speed_RPM: RotationalSpeedRPM,
    AnalogInputType.Current_Amps: CurrentAmps,
    AnalogInputType.Frequency_Hz: FrequencyHz,
    AnalogInputType.Power_Watts: PowerWatts,
    AnalogInputType.Power_Kilo_Watts: PowerKiloWatts,
    AnalogInputType.Energy_Kilo_Watt_Hours: EnergyKiloWattHours,
    AnalogInputType.Count: Count,
    AnalogInputType.Enthalpy_KJoules_Per_Kg: EnthalpyKJoulesPerKg,
    AnalogInputType.Time_Seconds: TimeSeconds,
}


class ApplicationType(t.IntStruct, t.uint32_t):
    # Index = Bits 0 to 15
    index: t.uint16_t
    # Type = Bits 16 to 23
    type: AnalogInputType
    # Group = Bits 24 to 31
    group: t.uint8_t



from __future__ import annotations

import zigpy.types as t

PROFILE_ID = 260


class DeviceType(t.enum16):
    # Generic
    ON_OFF_SWITCH = 0x0000
    LEVEL_CONTROL_SWITCH = 0x0001
    ON_OFF_OUTPUT = 0x0002
    LEVEL_CONTROLLABLE_OUTPUT = 0x0003
    SCENE_SELECTOR = 0x0004
    CONFIGURATION_TOOL = 0x0005
    REMOTE_CONTROL = 0x0006
    COMBINED_INTERFACE = 0x0007
    RANGE_EXTENDER = 0x0008
    MAIN_POWER_OUTLET = 0x0009
    DOOR_LOCK = 0x000A
    DOOR_LOCK_CONTROLLER = 0x000B
    SIMPLE_SENSOR = 0x000C
    CONSUMPTION_AWARENESS_DEVICE = 0x000D
    HOME_GATEWAY = 0x0050
    SMART_PLUG = 0x0051
    WHITE_GOODS = 0x0052
    METER_INTERFACE = 0x0053
    # Lighting
    ON_OFF_LIGHT = 0x0100
    DIMMABLE_LIGHT = 0x0101
    COLOR_DIMMABLE_LIGHT = 0x0102
    ON_OFF_LIGHT_SWITCH = 0x0103
    DIMMER_SWITCH = 0x0104
    COLOR_DIMMER_SWITCH = 0x0105
    LIGHT_SENSOR = 0x0106
    OCCUPANCY_SENSOR = 0x0107
    # ZLO device types
    ON_OFF_BALLAST = 0x0108
    DIMMABLE_BALLAST = 0x0109
    ON_OFF_PLUG_IN_UNIT = 0x010A
    DIMMABLE_PLUG_IN_UNIT = 0x010B
    COLOR_TEMPERATURE_LIGHT = 0x010C
    EXTENDED_COLOR_LIGHT = 0x010D
    LIGHT_LEVEL_SENSOR = 0x010E
    # Closure
    SHADE = 0x0200
    SHADE_CONTROLLER = 0x0201
    WINDOW_COVERING_DEVICE = 0x0202
    WINDOW_COVERING_CONTROLLER = 0x0203
    # HVAC
    HEATING_COOLING_UNIT = 0x0300
    THERMOSTAT = 0x0301
    TEMPERATURE_SENSOR = 0x0302
    PUMP = 0x0303
    PUMP_CONTROLLER = 0x0304
    PRESSURE_SENSOR = 0x0305
    FLOW_SENSOR = 0x0306
    MINI_SPLIT_AC = 0x0307
    # Intruder Alarm Systems
    IAS_CONTROL = 0x0400  # IAS Control and Indicating Equipment
    IAS_ANCILLARY_CONTROL = 0x0401  # IAS Ancillary Control Equipment
    IAS_ZONE = 0x0402
    IAS_WARNING_DEVICE = 0x0403
    # ZLO device types, continued
    COLOR_CONTROLLER = 0x0800
    COLOR_SCENE_CONTROLLER = 0x0810
    NON_COLOR_CONTROLLER = 0x0820
    NON_COLOR_SCENE_CONTROLLER = 0x0830
    CONTROL_BRIDGE = 0x0840
    ON_OFF_SENSOR = 0x0850


CLUSTERS = {
    # Generic
    DeviceType.ON_OFF_SWITCH: ([0x0007], [0x0004, 0x0005, 0x0006]),
    DeviceType.LEVEL_CONTROL_SWITCH: ([0x0007], [0x0004, 0x0005, 0x0006, 0x0008]),
    DeviceType.ON_OFF_OUTPUT: ([0x0004, 0x0005, 0x0006], []),
    DeviceType.LEVEL_CONTROLLABLE_OUTPUT: ([0x0004, 0x0005, 0x0006, 0x0008], []),
    DeviceType.SCENE_SELECTOR: ([], [0x0004, 0x0005]),
    DeviceType.REMOTE_CONTROL: ([], [0x0004, 0x0005, 0x0006, 0x0008]),
    DeviceType.MAIN_POWER_OUTLET: ([0x0004, 0x0005, 0x0006], []),
    DeviceType.SMART_PLUG: ([0x0004, 0x0005, 0x0006], []),
    # Lighting
    DeviceType.ON_OFF_LIGHT: ([0x0004, 0x0005, 0x0006, 0x0008], []),
    DeviceType.DIMMABLE_LIGHT: ([0x0004, 0x0005, 0x0006, 0x0008], []),
    DeviceType.COLOR_DIMMABLE_LIGHT: ([0x0004, 0x0005, 0x0006, 0x0008, 0x0300], []),
    DeviceType.ON_OFF_LIGHT_SWITCH: ([0x0007], [0x0004, 0x0005, 0x0006]),
    DeviceType.DIMMER_SWITCH: ([0x0007], [0x0004, 0x0005, 0x0006, 0x0008]),
    DeviceType.COLOR_DIMMER_SWITCH: (
        [0x0007],
        [0x0004, 0x0005, 0x0006, 0x0008, 0x0300],
    ),
    DeviceType.LIGHT_SENSOR: ([0x0400], []),
    DeviceType.OCCUPANCY_SENSOR: ([0x0406], []),
    DeviceType.COLOR_TEMPERATURE_LIGHT: (
        [0x0003, 0x0004, 0x0005, 0x0006, 0x0008, 0x0300],
        [],
    ),
    DeviceType.EXTENDED_COLOR_LIGHT: (
        [0x0003, 0x0004, 0x0005, 0x0006, 0x0008, 0x0300],
        [],
    ),
    # Closures
    DeviceType.WINDOW_COVERING_DEVICE: ([0x0004, 0x0005, 0x0102], []),
    # HVAC
    DeviceType.THERMOSTAT: ([0x0201, 0x0204], [0x0200, 0x0202, 0x0203]),
}



from __future__ import annotations

import asyncio
from datetime import UTC, datetime
import re
from typing import Any
from zoneinfo import ZoneInfo

import pytest

from zigpy import device, types, zcl
import zigpy.endpoint
from zigpy.ota import OTA, OtaImagesResult
from zigpy.zcl import OtaImageAvailableEvent, OtaQueryCacheUpdatedEvent, foundation
from zigpy.zcl.clusters.general import Basic, Ota, Time
import zigpy.zcl.clusters.security as sec
from zigpy.zdo import types as zdo_t

from .async_mock import AsyncMock, MagicMock, call, patch, sentinel

IMAGE_SIZE = 0x2345
IMAGE_OFFSET = 0x2000


def test_registry():
    for cluster_id, cluster in zcl.Cluster._registry.items():
        assert 0 <= getattr(cluster, "cluster_id", -1) <= 65535
        assert cluster_id == cluster.cluster_id
        assert issubclass(cluster, zcl.Cluster)


def test_attributes():
    for cluster in zcl.Cluster._registry.values():
        for attrid, attr in cluster.attributes.items():
            assert 0 <= attrid <= 0xFFFF
            assert isinstance(attr, zcl.foundation.ZCLAttributeDef)
            assert attr.id == attrid
            assert attr.name
            assert attr.type
            assert callable(attr.type.deserialize)
            assert callable(attr.type.serialize)


def _test_commands(cmdattr):
    for cluster in zcl.Cluster._registry.values():
        for cmdid, cmdspec in getattr(cluster, cmdattr).items():
            assert 0 <= cmdid <= 0xFF

            assert cmdspec.id == cmdid
            assert isinstance(cmdspec, zcl.foundation.ZCLCommandDef)
            assert issubclass(cmdspec.schema, types.Struct)

            for field in cmdspec.schema.fields:
                assert callable(field.type.deserialize)
                assert callable(field.type.serialize)


def test_server_commands():
    _test_commands("server_commands")


def test_client_commands():
    _test_commands("client_commands")


def test_ep_attributes():
    seen = set()
    for cluster in zcl.Cluster._registry.values():
        assert isinstance(cluster.ep_attribute, str)
        assert re.match(r"^[a-z_][a-z0-9_]*$", cluster.ep_attribute)
        assert cluster.ep_attribute not in seen
        seen.add(cluster.ep_attribute)

        ep = zigpy.endpoint.Endpoint(None, 1)
        assert not hasattr(ep, cluster.ep_attribute)


async def read_attributes(cluster, attribute_ids: list[int]) -> dict[int, Any]:
    schema = foundation.GENERAL_COMMANDS[
        foundation.GeneralCommand.Read_Attributes
    ].schema
    hdr, _ = cluster._create_request(
        general=True,
        command_id=foundation.GeneralCommand.Read_Attributes,
        schema=schema,
        disable_default_response=False,
        direction=foundation.Direction.Client_to_Server,
        args=(),
        kwargs={"attribute_ids": attribute_ids},
    )

    command = schema(attribute_ids=attribute_ids)

    with patch.object(cluster, "reply") as reply_mock:
        cluster.handle_message(hdr, command)
        call = reply_mock.mock_calls[0]

    return call.args[2](call.args[3])


async def test_basic_cluster():
    ep = MagicMock()
    ep.reply = AsyncMock()

    cluster = Basic(ep)

    rsp = await read_attributes(
        cluster,
        [
            Basic.AttributeDefs.zcl_version.id,
            Basic.AttributeDefs.power_source.id,
            Basic.AttributeDefs.serial_number.id,
        ],
    )

    assert rsp.status_records[0] == foundation.ReadAttributeRecord(
        attrid=Basic.AttributeDefs.zcl_version.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.uint8,
            value=8,
        ),
    )

    assert rsp.status_records[1] == foundation.ReadAttributeRecord(
        attrid=Basic.AttributeDefs.power_source.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.enum8,
            value=Basic.PowerSource.DC_Source,
        ),
    )

    assert rsp.status_records[2] == foundation.ReadAttributeRecord(
        attrid=Basic.AttributeDefs.serial_number.id,
        status=foundation.Status.UNSUPPORTED_ATTRIBUTE,
    )


def _make_patched_datetime(fake_now):
    """Create a PatchedDatetime class that returns `fake_now` as the current time."""

    class PatchedDatetime(datetime):
        _fake_now = fake_now

        def astimezone(self):
            return self.replace(tzinfo=self._fake_now.tzinfo)

        @classmethod
        def now(cls, tzinfo=None):
            if tzinfo is None:
                return cls(
                    cls._fake_now.year,
                    cls._fake_now.month,
                    cls._fake_now.day,
                    cls._fake_now.hour,
                    cls._fake_now.minute,
                    cls._fake_now.second,
                )
            else:
                assert tzinfo is UTC
                return (
                    cls(
                        cls._fake_now.year,
                        cls._fake_now.month,
                        cls._fake_now.day,
                        cls._fake_now.hour,
                        cls._fake_now.minute,
                        cls._fake_now.second,
                        tzinfo=tzinfo,
                    )
                    - cls._fake_now.utcoffset()
                )

    return PatchedDatetime


@pytest.mark.parametrize(
    ("fake_now", "expected_utc", "expected_tz", "expected_standard", "expected_local"),
    [
        pytest.param(
            # January: PST (UTC-8), no DST
            datetime(2000, 1, 2, 0, 0, 0, tzinfo=ZoneInfo("America/Los_Angeles")),
            # UTC time: midnight Jan 2 PST = 08:00 Jan 2 UTC = 1 day + 8 hours
            24 * 60 * 60 + 8 * 60 * 60,
            # Standard timezone offset: UTC-8
            -(8 * 60 * 60),
            # Standard time: Time + TimeZone = midnight Jan 2 local (no DST)
            24 * 60 * 60,
            # Local time: midnight Jan 2 = 1 day from epoch
            24 * 60 * 60,
            id="winter_no_dst",
        ),
        pytest.param(
            # July: PDT (UTC-7), DST active
            datetime(2000, 7, 2, 0, 0, 0, tzinfo=ZoneInfo("America/Los_Angeles")),
            # UTC time: midnight Jul 2 PDT = 07:00 Jul 2 UTC
            183 * 24 * 60 * 60 + 7 * 60 * 60,
            # Standard timezone offset: still UTC-8 (DST not included)
            -(8 * 60 * 60),
            # Standard time: Time + TimeZone = 11 PM Jul 1 standard (no DST adjustment)
            183 * 24 * 60 * 60 - 1 * 60 * 60,
            # Local time: midnight Jul 2 local
            183 * 24 * 60 * 60,
            id="summer_with_dst",
        ),
    ],
)
async def test_time_cluster(
    fake_now, expected_utc, expected_tz, expected_standard, expected_local
):
    ep = MagicMock()
    ep.reply = AsyncMock()

    cluster = Time(ep)

    Read_Attributes_rsp = foundation.GENERAL_COMMANDS[
        foundation.GeneralCommand.Read_Attributes_rsp
    ].schema

    PatchedDatetime = _make_patched_datetime(fake_now)

    with patch("zigpy.zcl.clusters.general.datetime", PatchedDatetime):
        # Supported attributes
        rsp1 = await read_attributes(
            cluster,
            [
                Time.AttributeDefs.time.id,
                Time.AttributeDefs.time_status.id,
                Time.AttributeDefs.time_zone.id,
                Time.AttributeDefs.standard_time.id,
                Time.AttributeDefs.local_time.id,
            ],
        )

    assert rsp1.status_records[0] == foundation.ReadAttributeRecord(
        attrid=Time.AttributeDefs.time.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.UTC,
            value=expected_utc,
        ),
    )

    assert rsp1.status_records[1] == foundation.ReadAttributeRecord(
        attrid=Time.AttributeDefs.time_status.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.map8,
            value=(Time.TimeStatus.Master | Time.TimeStatus.Master_for_Zone_and_DST),
        ),
    )

    assert rsp1.status_records[2] == foundation.ReadAttributeRecord(
        attrid=Time.AttributeDefs.time_zone.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.int32,
            value=expected_tz,
        ),
    )

    assert rsp1.status_records[3] == foundation.ReadAttributeRecord(
        attrid=Time.AttributeDefs.standard_time.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.uint32,
            value=expected_standard,
        ),
    )

    assert rsp1.status_records[4] == foundation.ReadAttributeRecord(
        attrid=Time.AttributeDefs.local_time.id,
        status=foundation.Status.SUCCESS,
        value=foundation.TypeValue(
            type=foundation.DataTypeId.uint32,
            value=expected_local,
        ),
    )

    # Unsupported
    rsp2 = await read_attributes(cluster, [0xABCD])
    assert rsp2 == Read_Attributes_rsp(
        status_records=[
            foundation.ReadAttributeRecord(
                attrid=0xABCD,
                status=foundation.Status.UNSUPPORTED_ATTRIBUTE,
            )
        ]
    )


@pytest.fixture
def dev(monkeypatch, app_mock):
    monkeypatch.setattr(device, "APS_REPLY_TIMEOUT_EXTENDED", 0.1)
    ieee = types.EUI64(map(types.uint8_t, [0, 1, 2, 3, 4, 5, 6, 7]))

    dev = device.Device(app_mock, ieee, 65535)
    node_desc = zdo_t.NodeDescriptor(1, 1, 1, 4, 5, 6, 7, 8)
    with patch.object(
        dev.zdo, "Node_Desc_req", new=AsyncMock(return_value=(0, 0xFFFF, node_desc))
    ):
        yield dev


@pytest.fixture
def ota_cluster(dev):
    ep = dev.add_endpoint(1)

    cluster = zcl.Cluster._registry[0x0019](ep)
    ep.in_clusters[cluster.cluster_id] = cluster

    with (
        patch.object(cluster, "reply", AsyncMock()),
        patch.object(cluster, "request", AsyncMock()),
    ):
        yield cluster


async def test_ota_handle_cluster_req_wrapper(ota_cluster, caplog):
    ota_cluster._handle_query_next_image = AsyncMock()

    hdr = zigpy.zcl.foundation.ZCLHeader.cluster(123, 0x01)
    ota_cluster.handle_cluster_request(hdr, [sentinel.args])
    assert ota_cluster._handle_query_next_image.call_count == 1
    assert ota_cluster._handle_query_next_image.mock_calls[0].args == (
        hdr,
        [sentinel.args],
    )
    ota_cluster._handle_query_next_image.reset_mock()

    # This command isn't currently handled
    hdr.command_id = 0x08
    ota_cluster.handle_cluster_request(hdr, [sentinel.just_args])
    assert ota_cluster._handle_query_next_image.call_count == 0


async def test_ota_handle_query_next_image(ota_cluster):
    dev = ota_cluster.endpoint.device

    ota_cluster.query_next_image_response = AsyncMock()
    dev.ota_in_progress = False

    # TODO: get rid of `sentinel` and mock the actual command
    hdr = zigpy.zcl.foundation.ZCLHeader.cluster(
        tsn=0x12, command_id=Ota.ServerCommandDefs.query_next_image.id
    )
    cmd = MagicMock()

    cache_events = []
    ota_cluster.on_event(OtaQueryCacheUpdatedEvent.event_type, cache_events.append)

    image_events = []
    ota_cluster.on_event(OtaImageAvailableEvent.event_type, image_events.append)

    # Use the real check_cluster_for_ota so it emits OtaImageAvailableEvent
    ota = dev.application.ota
    ota.check_cluster_for_ota = lambda c: OTA.check_cluster_for_ota(ota, c)

    # No image is available
    ota.get_ota_images = AsyncMock(
        return_value=OtaImagesResult(upgrades=(), downgrades=())
    )
    ota_cluster.handle_cluster_request(hdr, cmd)
    await asyncio.sleep(0)

    assert ota_cluster.query_next_image_response.mock_calls == [
        call(zcl.foundation.Status.NO_IMAGE_AVAILABLE, tsn=hdr.tsn)
    ]
    assert len(image_events) == 1
    assert isinstance(image_events[0], OtaImageAvailableEvent)
    assert image_events[0].images_result == OtaImagesResult(upgrades=(), downgrades=())
    assert image_events[0].query_cmd is cmd
    assert ota_cluster.last_query_cmd is cmd
    assert len(cache_events) == 1
    assert isinstance(cache_events[0], OtaQueryCacheUpdatedEvent)

    ota_cluster.query_next_image_response.reset_mock()
    image_events.clear()

    # Now one is available
    img = MagicMock()
    ota.get_ota_images = AsyncMock(
        return_value=OtaImagesResult(upgrades=(img,), downgrades=())
    )
    ota_cluster.handle_cluster_request(hdr, cmd)
    await asyncio.sleep(0)

    assert ota_cluster.query_next_image_response.mock_calls == [
        call(zcl.foundation.Status.NO_IMAGE_AVAILABLE, tsn=hdr.tsn)
    ]
    assert len(image_events) == 1
    assert image_events[0].images_result == OtaImagesResult(
        upgrades=(img,), downgrades=()
    )
    assert image_events[0].query_cmd is cmd


async def test_ota_handle_image_block_req(ota_cluster):
    dev = ota_cluster.endpoint.device

    ota_cluster.image_block_response = AsyncMock()
    dev.ota_in_progress = False

    hdr = zigpy.zcl.foundation.ZCLHeader.cluster(
        tsn=0x12, command_id=Ota.ServerCommandDefs.image_block.id
    )
    cmd = MagicMock()

    # Stop the upgrade, none is in progress
    ota_cluster.handle_cluster_request(hdr, cmd)
    await asyncio.sleep(0)

    assert ota_cluster.image_block_response.mock_calls == [
        call(zcl.foundation.Status.ABORT, tsn=hdr.tsn)
    ]

    ota_cluster.image_block_response.reset_mock()

    # If we flip the progress flag, send nothing
    dev.ota_in_progress = True
    ota_cluster.handle_cluster_request(hdr, cmd)
    await asyncio.sleep(0)

    assert ota_cluster.image_block_response.mock_calls == []


def test_ias_zone_type():
    extra = b"\xaa\x55"
    zone, rest = sec.IasZone.ZoneType.deserialize(b"\x0d\x00" + extra)
    assert rest == extra
    assert zone is sec.IasZone.ZoneType.Motion_Sensor

    zone, rest = sec.IasZone.ZoneType.deserialize(b"\x81\x81" + extra)
    assert rest == extra
    assert zone.name.startswith("manufacturer_specific")
    assert zone.value == 0x8181


def test_ias_ace_audible_notification():
    extra = b"\xaa\x55"
    notification_type, rest = sec.IasAce.AudibleNotification.deserialize(
        b"\x00" + extra
    )
    assert rest == extra
    assert notification_type is sec.IasAce.AudibleNotification.Mute

    notification_type, rest = sec.IasAce.AudibleNotification.deserialize(
        b"\x81" + extra
    )
    assert rest == extra
    assert notification_type.name.startswith("manufacturer_specific")
    assert notification_type.value == 0x81


def test_basic_cluster_power_source():
    extra = b"The rest of the owl\xaa\x55"
    pwr_src, rest = zcl.clusters.general.Basic.PowerSource.deserialize(b"\x81" + extra)

    assert rest == extra
    assert pwr_src == zcl.clusters.general.Basic.PowerSource.Mains_single_phase
    assert pwr_src == 0x01
    assert pwr_src.value == 0x01
    assert pwr_src.battery_backup


@pytest.mark.parametrize(
    ("raw", "mode", "name"),
    [
        (0x00, 0, "Stop"),
        (0x01, 0, "Stop"),
        (0x02, 0, "Stop"),
        (0x03, 0, "Stop"),
        (0x30, 3, "Emergency"),
        (0x31, 3, "Emergency"),
        (0x32, 3, "Emergency"),
        (0x33, 3, "Emergency"),
    ],
)
def test_security_iaswd_warning_mode(raw, mode, name):
    """Test warning command class of IasWD cluster."""

    def _test(warning, data):
        assert warning.serialize() == data
        assert warning == raw
        assert warning.mode == mode
        assert warning.mode.name == name
        warning.mode = mode
        assert warning.serialize() == data
        assert warning.mode == mode

    data = types.uint8_t(raw).serialize()
    _test(sec.IasWd.Warning(raw), data)

    extra = b"The rest of the owl\xaa\x55"
    warn, rest = sec.IasWd.Warning.deserialize(data + extra)
    assert rest == extra
    _test(warn, data)
    repr(warn)


def test_security_iaswd_warning_mode_2():
    """Test warning command class of IasWD cluster."""

    def _test(data, raw, mode, name):
        warning, _ = sec.IasWd.Warning.deserialize(data)
        assert warning.serialize() == data
        assert warning == raw
        assert warning.mode == mode
        assert warning.mode.name == name
        warning.mode = mode
        assert warning.serialize() == data
        assert warning.mode == mode

    for mode in sec.IasWd.Warning.WarningMode:
        for other in range(16):
            raw = mode << 4 | other
            data = types.uint8_t(raw).serialize()
            _test(data, raw, mode.value, mode.name)


def test_security_iaswd_warning_strobe():
    """Test strobe of warning command class of IasWD cluster."""

    for strobe in sec.IasWd.Warning.Strobe:
        for mode in range(16):
            for siren in range(4):
                raw = mode << 4 | siren
                raw |= strobe.value << 2
                data = types.uint8_t(raw).serialize()
                warning, _ = sec.IasWd.Warning.deserialize(data)
                assert warning.serialize() == data
                assert warning == raw
                assert warning.strobe == strobe.value
                assert warning.strobe.name == strobe.name
                warning.strobe = strobe
                assert warning.serialize() == data
                assert warning.strobe == strobe.value


def test_security_iaswd_warning_siren():
    """Test siren of warning command class of IasWD cluster."""

    for siren in sec.IasWd.Warning.SirenLevel:
        for mode in range(16):
            for strobe in range(4):
                raw = mode << 4 | (strobe << 2)
                raw |= siren.value
                data = types.uint8_t(raw).serialize()
                warning, _ = sec.IasWd.Warning.deserialize(data)
                assert warning.serialize() == data
                assert warning == raw
                assert warning.level == siren.value
                assert warning.level.name == siren.name
                warning.level = siren
                assert warning.serialize() == data
                assert warning.level == siren.value


@pytest.mark.parametrize(
    ("raw", "mode", "name"),
    [
        (0x00, 0, "Armed"),
        (0x01, 0, "Armed"),
        (0x02, 0, "Armed"),
        (0x03, 0, "Armed"),
        (0x10, 1, "Disarmed"),
        (0x11, 1, "Disarmed"),
        (0x12, 1, "Disarmed"),
        (0x13, 1, "Disarmed"),
    ],
)
def test_security_iaswd_squawk_mode(raw, mode, name):
    """Test squawk command class of IasWD cluster."""

    def _test(squawk, data):
        assert squawk.serialize() == data
        assert squawk == raw
        assert squawk.mode == mode
        assert squawk.mode.name == name
        squawk.mode = mode
        assert squawk.serialize() == data
        assert squawk.mode == mode

    data = types.uint8_t(raw).serialize()
    _test(sec.IasWd.Squawk(raw), data)

    extra = b"The rest of the owl\xaa\x55"
    squawk, rest = sec.IasWd.Squawk.deserialize(data + extra)
    assert rest == extra
    _test(squawk, data)
    repr(squawk)


def test_security_iaswd_squawk_strobe():
    """Test strobe of squawk command class of IasWD cluster."""

    for strobe in sec.IasWd.Squawk.Strobe:
        for mode in range(16):
            for level in range(4):
                raw = mode << 4 | level
                raw |= strobe.value << 3
                data = types.uint8_t(raw).serialize()
                squawk, _ = sec.IasWd.Squawk.deserialize(data)
                assert squawk.serialize() == data
                assert squawk == raw
                assert squawk.strobe == strobe.value
                assert squawk.strobe == strobe
                assert squawk.strobe.name == strobe.name
                squawk.strobe = strobe
                assert squawk.serialize() == data
                assert squawk.strobe == strobe


def test_security_iaswd_squawk_level():
    """Test level of squawk command class of IasWD cluster."""

    for level in sec.IasWd.Squawk.SquawkLevel:
        for other in range(64):
            raw = other << 2 | level.value
            data = types.uint8_t(raw).serialize()
            squawk, _ = sec.IasWd.Squawk.deserialize(data)
            assert squawk.serialize() == data
            assert squawk == raw
            assert squawk.level == level.value
            assert squawk.level == level
            assert squawk.level.name == level.name
            squawk.level = level
            assert squawk.serialize() == data
            assert squawk.level == level


def test_hvac_thermostat_system_type():
    """Test system_type class."""

    hvac = zcl.clusters.hvac
    sys_type = hvac.Thermostat.SystemType(0x00)
    assert sys_type.cooling_system_stage == hvac.CoolingSystemStage.Cool_Stage_1
    assert sys_type.heating_system_stage == hvac.HeatingSystemStage.Heat_Stage_1
    assert sys_type.heating_fuel_source == hvac.HeatingFuelSource.Electric
    assert sys_type.heating_system_type == hvac.HeatingSystemType.Conventional

    sys_type = hvac.Thermostat.SystemType(0x35)
    assert sys_type.cooling_system_stage == hvac.CoolingSystemStage.Cool_Stage_2
    assert sys_type.heating_system_stage == hvac.HeatingSystemStage.Heat_Stage_2
    assert sys_type.heating_fuel_source == hvac.HeatingFuelSource.Gas
    assert sys_type.heating_system_type == hvac.HeatingSystemType.Heat_Pump


@patch("zigpy.zcl.Cluster.send_default_rsp")
async def test_ias_zone(send_rsp_mock):
    """Test sending default response on zone status notification."""

    ep = MagicMock()
    ep.reply = AsyncMock()
    t = zcl.Cluster._registry[sec.IasZone.cluster_id](ep, is_server=False)

    # suppress default response
    hdr, args = t.deserialize(b"\tK\x00&\x00\x00\x00\x00\x00")
    hdr.frame_control = hdr.frame_control.replace(disable_default_response=True)
    t.handle_message(hdr, args)
    assert send_rsp_mock.call_count == 0
    send_rsp_mock.reset_mock()

    # this should generate a default response
    hdr.frame_control = hdr.frame_control.replace(disable_default_response=False)
    t.handle_message(hdr, args)
    assert send_rsp_mock.call_count == 1
    send_rsp_mock.reset_mock()

    t = zcl.Cluster._registry[sec.IasZone.cluster_id](ep, is_server=True)

    # suppress default response
    hdr, args = t.deserialize(b"\tK\x00&\x00\x00\x00\x00\x00")
    hdr.frame_control = hdr.frame_control.replace(disable_default_response=True)
    t.handle_message(hdr, args)
    assert send_rsp_mock.call_count == 0
    send_rsp_mock.reset_mock()

    # this should generate a default response
    hdr.frame_control = hdr.frame_control.replace(disable_default_response=False)
    t.handle_message(hdr, args)
    assert send_rsp_mock.call_count == 1
    send_rsp_mock.reset_mock()


def test_ota_image_block_field_control():
    """Test OTA image_block with field control deserializes properly."""
    data = bytes.fromhex("01d403020b101d01001f000100000000400000")

    ep = MagicMock()
    cluster = zcl.clusters.general.Ota(ep)

    hdr, response = cluster.deserialize(data)
    assert hdr.serialize() + response.serialize() == data

    image_block = cluster.commands_by_name["image_block"].schema

    assert response == image_block(
        field_control=image_block.FieldControl.MinimumBlockPeriod,
        manufacturer_code=4107,
        image_type=285,
        file_version=0x01001F00,
        file_offset=0,
        maximum_data_size=64,
        minimum_block_period=0,
    )

    assert response.request_node_addr is None


def test_general_analog_in_application_type():
    """Test AnalogInput General Cluster, Application Type Attribute."""
    app_type = zcl.clusters.general_const.ApplicationType(
        0x00_01_0007
    )  # Group 0x00, Type 0x01, Application 0x0007

    assert app_type.group == 0x00
    assert (
        app_type.type
        == zcl.clusters.general_const.AnalogInputType.Relative_Humidity_Percent
    )
    assert (
        app_type.index
        == zcl.clusters.general_const.RelativeHumidityPercent.Space_Humidity
    )


def test_ias_zone_enum_subclass_zcl_type():
    """Test `IasZone.zone_type` is an enum16, not a uint16."""

    assert sec.IasZone.AttributeDefs.zone_type.zcl_type == foundation.DataTypeId.enum16



"""Security and Safety Functional Domain"""

from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster, foundation
from zigpy.zcl.foundation import (
    BaseAttributeDefs,
    BaseCommandDefs,
    ZCLAttributeDef,
    ZCLCommandDef,
)


class ZoneState(t.enum8):
    Not_enrolled = 0x00
    Enrolled = 0x01


class ZoneType(t.enum16):
    """Zone type enum."""

    @classmethod
    def _missing_(cls, value):
        new = cls._member_type_.__new__(cls, value)
        new._name_ = f"manufacturer_specific_{new._hex_repr().lower()}"
        new._value_ = value
        return new

    Standard_CIE = 0x0000
    Motion_Sensor = 0x000D
    Contact_Switch = 0x0015
    Fire_Sensor = 0x0028
    Water_Sensor = 0x002A
    Carbon_Monoxide_Sensor = 0x002B
    Personal_Emergency_Device = 0x002C
    Vibration_Movement_Sensor = 0x002D
    Remote_Control = 0x010F
    Key_Fob = 0x0115
    Key_Pad = 0x021D
    Standard_Warning_Device = 0x0225
    Glass_Break_Sensor = 0x0226
    Security_Repeater = 0x0229
    Invalid_Zone_Type = 0xFFFF


class ZoneStatus(t.bitmap16):
    """ZoneStatus attribute."""

    Alarm_1 = 0x0001
    Alarm_2 = 0x0002
    Tamper = 0x0004
    Battery = 0x0008
    Supervision_reports = 0x0010
    Restore_reports = 0x0020
    Trouble = 0x0040
    AC_mains = 0x0080
    Test = 0x0100
    Battery_Defect = 0x0200


class EnrollResponse(t.enum8):
    """Enroll response code."""

    Success = 0x00
    Not_supported = 0x01
    No_enroll_permit = 0x02
    Too_many_zones = 0x03


class IasZone(Cluster):
    """The IAS Zone cluster defines an interface to the functionality of an IAS
    security zone device. IAS Zone supports up to two alarm types per zone, low battery
    reports and supervision of the IAS network.
    """

    cluster_id: Final[t.uint16_t] = 0x0500
    name: Final = "IAS Zone"
    ep_attribute: Final = "ias_zone"

    class AttributeDefs(BaseAttributeDefs):
        # Zone Information
        zone_state: Final = ZCLAttributeDef(
            id=0x0000, type=ZoneState, access="r", mandatory=True
        )
        zone_type: Final = ZCLAttributeDef(
            id=0x0001, type=ZoneType, access="r", mandatory=True
        )
        zone_status: Final = ZCLAttributeDef(
            id=0x0002, type=ZoneStatus, access="r", mandatory=True
        )
        # Zone Settings
        cie_addr: Final = ZCLAttributeDef(
            id=0x0010, type=t.EUI64, access="rw", mandatory=True
        )
        zone_id: Final = ZCLAttributeDef(
            id=0x0011, type=t.uint8_t, access="r", mandatory=True
        )
        # Both attributes will be supported/unsupported
        num_zone_sensitivity_levels_supported: Final = ZCLAttributeDef(
            id=0x0012, type=t.uint8_t, access="r"
        )
        current_zone_sensitivity_level: Final = ZCLAttributeDef(
            id=0x0013, type=t.uint8_t, access="rw"
        )

    class ServerCommandDefs(BaseCommandDefs):
        enroll_response: Final = ZCLCommandDef(
            id=0x00,
            schema={"enroll_response_code": EnrollResponse, "zone_id": t.uint8_t},
        )
        init_normal_op_mode: Final = ZCLCommandDef(
            id=0x01,
            schema={},
        )
        init_test_mode: Final = ZCLCommandDef(
            id=0x02,
            schema={
                "test_mode_duration": t.uint8_t,
                "current_zone_sensitivity_level": t.uint8_t,
            },
        )

    class ClientCommandDefs(BaseCommandDefs):
        status_change_notification: Final = ZCLCommandDef(
            id=0x00,
            schema={
                "zone_status": ZoneStatus,
                "extended_status": t.bitmap8,
                "zone_id": t.uint8_t,
                "delay": t.uint16_t,
            },
        )
        enroll: Final = ZCLCommandDef(
            id=0x01,
            schema={"zone_type": ZoneType, "manufacturer_code": t.uint16_t},
        )

    # For backwards compatibility
    ZoneState: Final = ZoneState
    ZoneType: Final = ZoneType
    ZoneStatus: Final = ZoneStatus
    EnrollResponse: Final = EnrollResponse


class AlarmStatus(t.enum8):
    """IAS ACE alarm status enum."""

    No_Alarm = 0x00
    Burglar = 0x01
    Fire = 0x02
    Emergency = 0x03
    Police_Panic = 0x04
    Fire_Panic = 0x05
    Emergency_Panic = 0x06


class ArmMode(t.enum8):
    """IAS ACE arm mode enum."""

    Disarm = 0x00
    Arm_Day_Home_Only = 0x01
    Arm_Night_Sleep_Only = 0x02
    Arm_All_Zones = 0x03


class ArmNotification(t.enum8):
    """IAS ACE arm notification enum."""

    All_Zones_Disarmed = 0x00
    Only_Day_Home_Zones_Armed = 0x01
    Only_Night_Sleep_Zones_Armed = 0x02
    All_Zones_Armed = 0x03
    Invalid_Arm_Disarm_Code = 0x04
    Not_Ready_To_Arm = 0x05
    Already_Disarmed = 0x06


class AudibleNotification(t.enum8):
    """IAS ACE audible notification enum."""

    @classmethod
    def _missing_(cls, value):
        new = cls._member_type_.__new__(cls, value)
        new._name_ = f"manufacturer_specific_{new._hex_repr().lower()}"
        new._value_ = value
        return new

    Mute = 0x00
    Default_Sound = 0x01


class BypassResponse(t.enum8):
    """Bypass result."""

    Zone_bypassed = 0x00
    Zone_not_bypassed = 0x01
    Not_allowed = 0x02
    Invalid_Zone_ID = 0x03
    Unknown_Zone_ID = 0x04
    Invalid_Code = 0x05


class PanelStatus(t.enum8):
    """IAS ACE panel status enum."""

    Panel_Disarmed = 0x00
    Armed_Stay = 0x01
    Armed_Night = 0x02
    Armed_Away = 0x03
    Exit_Delay = 0x04
    Entry_Delay = 0x05
    Not_Ready_To_Arm = 0x06
    In_Alarm = 0x07
    Arming_Stay = 0x08
    Arming_Night = 0x09
    Arming_Away = 0x0A


class ZoneStatusRsp(t.Struct):
    """Zone status response."""

    zone_id: t.uint8_t
    zone_status: ZoneStatus


class IasAce(Cluster):
    """IAS Ancillary Control Equipment cluster."""

    AlarmStatus: Final = AlarmStatus
    ArmMode: Final = ArmMode
    ArmNotification: Final = ArmNotification
    AudibleNotification: Final = AudibleNotification
    BypassResponse: Final = BypassResponse
    PanelStatus: Final = PanelStatus
    ZoneType: Final = IasZone.ZoneType
    ZoneStatus: Final = IasZone.ZoneStatus
    ZoneStatusRsp: Final = ZoneStatusRsp

    cluster_id: Final[t.uint16_t] = 0x0501
    name: Final = "IAS Ancillary Control Equipment"
    ep_attribute: Final = "ias_ace"

    class AttributeDefs(BaseAttributeDefs):
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        arm: Final = ZCLCommandDef(
            id=0x00,
            schema={
                "arm_mode": ArmMode,
                "arm_disarm_code": t.CharacterString,
                "zone_id": t.uint8_t,
            },
        )
        bypass: Final = ZCLCommandDef(
            id=0x01,
            schema={
                "zones_ids": t.LVList[t.uint8_t, t.uint8_t],
                "arm_disarm_code": t.CharacterString,
            },
        )
        emergency: Final = ZCLCommandDef(
            id=0x02,
            schema={},
        )
        fire: Final = ZCLCommandDef(
            id=0x03,
            schema={},
        )
        panic: Final = ZCLCommandDef(
            id=0x04,
            schema={},
        )
        get_zone_id_map: Final = ZCLCommandDef(
            id=0x05,
            schema={},
        )
        get_zone_info: Final = ZCLCommandDef(
            id=0x06,
            schema={"zone_id": t.uint8_t},
        )
        get_panel_status: Final = ZCLCommandDef(
            id=0x07,
            schema={},
        )
        get_bypassed_zone_list: Final = ZCLCommandDef(
            id=0x08,
            schema={},
        )
        get_zone_status: Final = ZCLCommandDef(
            id=0x09,
            schema={
                "starting_zone_id": t.uint8_t,
                "max_num_zone_ids": t.uint8_t,
                "zone_status_mask_flag": t.Bool,
                "zone_status_mask": ZoneStatus,
            },
        )

    class ClientCommandDefs(BaseCommandDefs):
        arm_response: Final = ZCLCommandDef(
            id=0x00,
            schema={"arm_notification": ArmNotification},
        )
        get_zone_id_map_response: Final = ZCLCommandDef(
            id=0x01,
            schema={"zone_id_map_sections": t.List[t.bitmap16]},
        )
        get_zone_info_response: Final = ZCLCommandDef(
            id=0x02,
            schema={
                "zone_id": t.uint8_t,
                "zone_type": ZoneType,
                "ieee": t.EUI64,
                "zone_label": t.CharacterString,
            },
        )
        zone_status_changed: Final = ZCLCommandDef(
            id=0x03,
            schema={
                "zone_id": t.uint8_t,
                "zone_status": ZoneStatus,
                "audible_notification": AudibleNotification,
                "zone_label": t.CharacterString,
            },
        )
        panel_status_changed: Final = ZCLCommandDef(
            id=0x04,
            schema={
                "panel_status": PanelStatus,
                "seconds_remaining": t.uint8_t,
                "audible_notification": AudibleNotification,
                "alarm_status": AlarmStatus,
            },
        )
        panel_status_response: Final = ZCLCommandDef(
            id=0x05,
            schema={
                "panel_status": PanelStatus,
                "seconds_remaining": t.uint8_t,
                "audible_notification": AudibleNotification,
                "alarm_status": AlarmStatus,
            },
        )
        set_bypassed_zone_list: Final = ZCLCommandDef(
            id=0x06,
            schema={"zone_ids": t.LVList[t.uint8_t, t.uint8_t]},
        )
        bypass_response: Final = ZCLCommandDef(
            id=0x07,
            schema={"bypass_results": t.LVList[BypassResponse, t.uint8_t]},
        )
        get_zone_status_response: Final = ZCLCommandDef(
            id=0x08,
            schema={
                "zone_status_complete": t.Bool,
                "zone_statuses": t.LVList[ZoneStatusRsp, t.uint8_t],
            },
        )


class Strobe(t.enum8):
    No_strobe = 0x00
    Strobe = 0x01


class _SquawkOrWarningCommand:  # noqa: PLW1641
    def __init__(self, value: int = 0) -> None:
        self.value = t.uint8_t(value)

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[_SquawkOrWarningCommand, bytes]:
        val, data = t.uint8_t.deserialize(data)
        return cls(val), data

    def serialize(self) -> bytes:
        return t.uint8_t(self.value).serialize()

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}.mode={self.mode.name} "
            f"strobe={self.strobe.name} level={self.level.name}: "
            f"{self.value}>"
        )

    def __eq__(self, other):
        """Compare to int."""
        return self.value == other


class StrobeLevel(t.enum8):
    Low_level_strobe = 0x00
    Medium_level_strobe = 0x01
    High_level_strobe = 0x02
    Very_high_level_strobe = 0x03


class SirenLevel(t.enum8):
    Low_level_sound = 0x00
    Medium_level_sound = 0x01
    High_level_sound = 0x02
    Very_high_level_sound = 0x03


class WarningMode(t.enum8):
    Stop = 0x00
    Burglar = 0x01
    Fire = 0x02
    Emergency = 0x03
    Police_Panic = 0x04
    Fire_Panic = 0x05
    Emergency_Panic = 0x06


class WarningType(_SquawkOrWarningCommand):
    @property
    def mode(self) -> WarningMode:
        return WarningMode((self.value >> 4) & 0x0F)

    @mode.setter
    def mode(self, mode: WarningMode) -> None:
        self.value = (self.value & 0xF) | (mode << 4)

    @property
    def strobe(self) -> Strobe:
        return Strobe((self.value >> 2) & 0x01)

    @strobe.setter
    def strobe(self, strobe: Strobe) -> None:
        self.value = (self.value & 0xF7) | (
            (strobe & 0x01) << 2  # type:ignore[operator]
        )

    @property
    def level(self) -> SirenLevel:
        return SirenLevel(self.value & 0x03)

    @level.setter
    def level(self, level: SirenLevel) -> None:
        self.value = (self.value & 0xFC) | (level & 0x03)

    # For backwards compatibility
    Strobe: Final = Strobe
    WarningMode: Final = WarningMode
    SirenLevel: Final = SirenLevel


class SquawkLevel(t.enum8):
    Low_level_sound = 0x00
    Medium_level_sound = 0x01
    High_level_sound = 0x02
    Very_high_level_sound = 0x03


class SquawkMode(t.enum8):
    Armed = 0x00
    Disarmed = 0x01


class Squawk(_SquawkOrWarningCommand):
    @property
    def mode(self) -> SquawkMode:
        return SquawkMode((self.value >> 4) & 0x0F)

    @mode.setter
    def mode(self, mode: SquawkMode) -> None:
        self.value = (self.value & 0xF) | ((mode & 0x0F) << 4)

    @property
    def strobe(self) -> Strobe:
        return Strobe((self.value >> 3) & 0x01)

    @strobe.setter
    def strobe(self, strobe: Strobe) -> None:
        self.value = (self.value & 0xF7) | (strobe << 3)  # type:ignore[operator]

    @property
    def level(self) -> SquawkLevel:
        return SquawkLevel(self.value & 0x03)

    @level.setter
    def level(self, level: SquawkLevel) -> None:
        self.value = (self.value & 0xFC) | (level & 0x03)

    # For backwards compatibility
    Strobe: Final = Strobe
    SquawkLevel: Final = SquawkLevel
    SquawkMode: Final = SquawkMode


class IasWd(Cluster):
    """The IAS WD cluster provides an interface to the functionality of any Warning
    Device equipment of the IAS system. Using this cluster, a Zigbee enabled CIE device
    can access a Zigbee enabled IAS WD device and issue alarm warning indications
    (siren, strobe lighting, etc.) when a system alarm condition is detected
    """

    StrobeLevel: Final = StrobeLevel
    Warning: Final = WarningType
    Squawk: Final = Squawk

    cluster_id: Final[t.uint16_t] = 0x0502
    name: Final = "IAS Warning Device"
    ep_attribute: Final = "ias_wd"

    class AttributeDefs(BaseAttributeDefs):
        max_duration: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rw", mandatory=True
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        start_warning: Final = ZCLCommandDef(
            id=0x00,
            schema={
                "warning": WarningType,
                "warning_duration": t.uint16_t,
                "strobe_duty_cycle": t.uint8_t,
                "stobe_level": StrobeLevel,
            },
        )
        squawk: Final = ZCLCommandDef(
            id=0x01,
            schema={"squawk": Squawk},
        )



from __future__ import annotations

from typing import Final

from zigpy.zcl import Cluster


class ManufacturerSpecificCluster(Cluster):
    cluster_id_range = (0xFC00, 0xFFFF)
    ep_attribute: Final = "manufacturer_specific"
    name: Final = "Manufacturer Specific"
